/* {\Tt all.h} for \impcore 23 */
/*UNIX -- FAS gcc barfs if stdio.h precedes stdarg.h */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <ctype.h>

/* shared type definitions 25e */
typedef struct Name Name;
typedef struct Namelist Namelist;   /* list of Name* */
/* shared type definitions 28d */
typedef struct Reader Reader;
/* shared type definitions 29a */
typedef struct Topreader Topreader;
/* shared type definitions 31e */
typedef void Printer(va_list*);
/* shared type definitions 499d */
typedef struct Parlist Parlist; /* list of Par* */
/* shared type definitions */
typedef struct Par Par;
/* type definitions for \impcore 25c */
typedef struct Explist Explist; /* list of Exp* */
/* type definitions for \impcore 26b */
typedef int Value;
typedef struct Valuelist Valuelist;     /* list of Value */
/* type definitions for \impcore 26d */
typedef struct Funlist Funlist; /* list of Fun */
/* type definitions for \impcore 27d */
typedef struct Valenv Valenv;
typedef struct Funenv Funenv;
/* type definitions for \impcore */
typedef struct Exp Exp;
/* type definitions for \impcore */
typedef struct Userfun Userfun;
typedef struct Toplevel Toplevel;
/* type definitions for \impcore */
typedef struct Fun Fun;
/* shared structure definitions 503a */
struct Reader {
    char *buf;               /* holds the last line read */
    int nbuf;                /* size of buf */
    int line;                /* current line number */
    const char *readername;  /* identifies this reader */

    FILE *fin;               /* filereader */
    const char *s;           /* stringreader */
};
/* shared structure definitions */
typedef enum { ATOM, LIST } Party;

struct Par {
    Party ty;
    union {
        Name *atom;
        Parlist *list;
    } u;
};

/* structure definitions for \impcore */
/* structure definitions for \impcore */
typedef enum { LITERAL, VAR, SET, IFX, WHILEX, BEGIN, APPLY } Expty;

struct Exp {
    Expty ty;
    union {
        Value literal;
        Name *var;
        struct {  Name *name; Exp *exp; } set;
        struct {  Exp *cond; Exp *true; Exp *false; } ifx;
        struct {  Exp *cond; Exp *exp; } whilex;
        Explist *begin;
        struct {  Name *name; Explist *actuals; } apply;
    } u;
};

/* structure definitions for \impcore */
struct Userfun {
    Namelist *formals;
    Exp *body;
};
typedef enum { EXP, VAL, DEFINE, USE } Toplevelty;

struct Toplevel {
    Toplevelty ty;
    union {
        Exp *exp;
        struct {  Name *name; Exp *exp; } val;
        struct {  Name *name; Userfun userfun; } define;
        Name *use;
    } u;
};

/* structure definitions for \impcore */
typedef enum { USERDEF, PRIMITIVE } Funty;

struct Fun {
    Funty ty;
    union {
        Userfun userdef;
        Name *primitive;
    } u;
};

/* structure definitions for \impcore */
struct Explist {
    Exp *hd;
    Explist *tl;
};

struct Parlist {
    Par *hd;
    Parlist *tl;
};

struct Valuelist {
    Value hd;
    Valuelist *tl;
};

struct Funlist {
    Fun hd;
    Funlist *tl;
};

struct Namelist {
    Name *hd;
    Namelist *tl;
};

/* shared function prototypes 25f */
Name *strtoname(const char *s);
const char *nametostr(Name *n);
/* shared function prototypes 26a */
Printer printname;
/* shared function prototypes 28e */
char *readline(Reader *r, char *prompt);
/* shared function prototypes 28f */
Reader *stringreader(const char *stringname, const char *s);
Reader *filereader  (const char *filename,   FILE *fin);
/* shared function prototypes 29b */
Toplevel *readtop(Topreader *r);
/* shared function prototypes 29c */
Topreader *topreader(Reader *r, int doprompt);
/* shared function prototypes 31d */
void print (const char *fmt, ...);
void vprint(const char *fmt, va_list va);
/* shared function prototypes 31f */
void installprinter(unsigned char c, Printer *fmt);
/* shared function prototypes 32a */
Printer printpercent, printstring, printdecimal;
/* shared function prototypes 32b */
void error(const char *fmt, ...);
extern jmp_buf errorjmp;        /* longjmp here on error */
/* shared function prototypes 32c */
void checkargc(Exp *e, int expected, int actual);
/* shared function prototypes 32d */
Name *duplicatename(Namelist *names);
/* shared function prototypes 499e */
/* shared function prototypes 500a */
Printer printpar;
/* shared function prototypes */
Parlist *readparlist(Reader *r, int doquote, int doprompt);
/* shared function prototypes */
Par *mkAtom(Name*);
Par *mkList(Parlist*);
/* function prototypes for \impcore 25d */
/* function prototypes for \impcore 26c */
Printer printexp, printtoplevel;
/* function prototypes for \impcore 27c */
Printer printvalue;
/* function prototypes for \impcore 27e */
/* function prototypes for \impcore 28a */
Printer printfun;
/* function prototypes for \impcore 28b */
Valenv *mkValenv(Namelist *vars, Valuelist *vals);
Funenv *mkFunenv(Namelist *vars, Funlist   *defs);
/* function prototypes for \impcore 28c */
Value fetchval(Name *name, Valenv *env);
Fun   fetchfun(Name *name, Funenv *env);
/* function prototypes for \impcore 29d */
int isvalbound(Name *name, Valenv *env);
int isfunbound(Name *name, Funenv *env);
/* function prototypes for \impcore 29e */
void bindval(Name *name, Value val, Valenv *env);
void bindfun(Name *name, Fun   fun, Funenv *env);
/* function prototypes for \impcore */
void topeval(Toplevel *t, Valenv *globals, Funenv *functions, int echo);
Value eval(Exp *e, Valenv *globals, Funenv *functions, Valenv *formals);
/* function prototypes for \impcore */
void readevalprint(Topreader *r, Valenv *globals, Funenv *functions, int echo);
/* function prototypes for \impcore */
/* function prototypes for \impcore */
Exp *mkLiteral(Value);
Exp *mkVar(Name*);
Exp *mkSet(Name *name, Exp *exp);
Exp *mkIfx(Exp *cond, Exp *true, Exp *false);
Exp *mkWhilex(Exp *cond, Exp *exp);
Exp *mkBegin(Explist*);
Exp *mkApply(Name *name, Explist *actuals);
/* function prototypes for \impcore */
Userfun mkUserfun(Namelist *formals, Exp *body);
Toplevel *mkExp(Exp*);
Toplevel *mkVal(Name *name, Exp *exp);
Toplevel *mkDefine(Name *name, Userfun userfun);
Toplevel *mkUse(Name*);
/* function prototypes for \impcore */
Fun mkUserdef(Userfun);
Fun mkPrimitive(Name*);
/* function prototypes for \impcore */
int      lengthEL (Explist *l);
Exp*     nthEL    (Explist *l, unsigned n);
Explist *mkEL     (Exp *hd, Explist *tl);
Printer  printexplist;
/* function prototypes for \impcore */
int      lengthPL (Parlist *l);
Par*     nthPL    (Parlist *l, unsigned n);
Parlist *mkPL     (Par *hd, Parlist *tl);
Printer  printparlist;
/* function prototypes for \impcore */
int        lengthVL (Valuelist *l);
Value      nthVL    (Valuelist *l, unsigned n);
Valuelist *mkVL     (Value hd, Valuelist *tl);
Printer    printvaluelist;
/* function prototypes for \impcore */
int      lengthFL (Funlist *l);
Fun      nthFL    (Funlist *l, unsigned n);
Funlist *mkFL     (Fun hd, Funlist *tl);
Printer  printfunlist;
/* function prototypes for \impcore */
int       lengthNL (Namelist *l);
Name*     nthNL    (Namelist *l, unsigned n);
Namelist *mkNL     (Name *hd, Namelist *tl);
Printer   printnamelist;
