/* env.c 45b */
#include "all.h"
/* env.c 45c */
struct Valenv {
    Namelist *nl;
    Valuelist *vl;
};
/* env.c 45d */
Valenv*
mkValenv(Namelist *nl, Valuelist *vl)
{
    Valenv *e;

    assert(lengthNL(nl) == lengthVL(vl));
    e = malloc(sizeof(*e));
    assert(e != NULL);
    e->nl = nl;
    e->vl = vl;
    return e;
}
/* env.c 46a */
static Value*
findval(Name *name, Valenv *env)
{
    Namelist *nl;
    Valuelist *vl;

    for (nl=env->nl, vl=env->vl; nl && vl; nl=nl->tl, vl=vl->tl)
        if (name == nl->hd)
            return &vl->hd;
    return NULL;
}
/* env.c 46b */
int
isvalbound(Name *name, Valenv *env)
{
    return findval(name, env) != NULL;
}
/* env.c 46c */
Value
fetchval(Name *name, Valenv *env)
{
    Value *vp;

    vp = findval(name, env);
    assert(vp != NULL);
    return *vp;
}
/* env.c 47 */
void
bindval(Name *name, Value val, Valenv *env)
{
    Value *vp;

    vp = findval(name, env);
    if (vp != NULL)
        *vp = val;              /* safe optimization */
    else {
        env->nl = mkNL(name, env->nl);
        env->vl = mkVL(val,  env->vl);
    }
}
/* env.c 505b */
struct Funenv {
    Namelist *nl;
    Funlist *fl;
};
/* env.c 505c */
Funenv*
mkFunenv(Namelist *nl, Funlist *fl)
{
    Funenv *e;

    assert(lengthNL(nl) == lengthFL(fl));
    e = malloc(sizeof *e);
    assert(e != NULL);
    e->nl = nl;
    e->fl = fl;
    return e;
}
/* env.c 506a */
static Fun*
findfun(Name *name, Funenv *env)
{
    Namelist *nl;
    Funlist *fl;

    for (nl = env->nl, fl = env->fl; nl && fl; nl = nl->tl, fl = fl->tl)
        if (name == nl->hd)
            return &fl->hd;
    return NULL;
}
/* env.c 506b */
int
isfunbound(Name *name, Funenv *env)
{
    return findfun(name, env) != NULL;
}
/* env.c 506c */
Fun
fetchfun(Name *name, Funenv *env)
{
    Fun *fp;

    fp = findfun(name, env);
    assert(fp != NULL);
    return *fp;
}
/* env.c 506d */
void
bindfun(Name *name, Fun fun, Funenv *env)
{
    Fun *fp;

    fp = findfun(name, env);
    if (fp != NULL)
        *fp = fun;              /* safe optimization */
    else {
        env->nl = mkNL(name, env->nl);
        env->fl = mkFL(fun,  env->fl);
    }
}
