/* parse.c 510c */
#include "all.h"

/* parse.c declarations 511d */
static Namelist *getnamelist(Name *f, Par *p, Parlist *pl);
/* parse.c declarations 513b */
static Exp *parseexp(Par*);
/* parse.c 511a */
static Toplevel*
parse(Par *p)
{
    switch (p->ty) {
    case ATOM:
        /* parse [[atom]] and return the result 511b */
        return mkExp(parseexp(p));
    case LIST:
        /* parse [[list]] and return the result 511c */
        {
            Parlist *pl;
            Name *first;

            pl = p->u.list;
            if (pl == NULL)
                error("%p: empty list", p);
            if (nthPL(pl, 0)->ty != ATOM)
                error("%p: first item of list not name", p);

            first = nthPL(pl, 0)->u.atom;
            if (first == strtoname("define")) {
                /* parse [[define]] and return the result 512b */
                Namelist *formals;
                Exp *body;
                Name *name;

                if (lengthPL(pl) != 4 || nthPL(pl, 1)->ty != ATOM || nthPL(pl, 2
                                                                  )->ty != LIST)
                    error("%p: usage: (define fun (formals) body)", p);

                name    = nthPL(pl, 1)->u.atom;
                formals = getnamelist(name, p, nthPL(pl, 2)->u.list);
                body    = parseexp(nthPL(pl, 3));
                return    mkDefine(name, mkUserfun(formals, body));
            }
            if (first == strtoname("val")) {
                /* parse [[val]] and return the result 512c */
                Exp *var, *exp;

                if (lengthPL(pl) != 3)
                    error("%p: usage: (val var exp)", p);

                var = parseexp(nthPL(pl, 1));
                if (var->ty != VAR)
                    error("%p: usage: (val var exp) (bad variable)", p);

                exp = parseexp(nthPL(pl, 2));
                return mkVal(var->u.var, exp);
            }
            if (first == strtoname("use")) {
                /* parse [[use]] and return the result 512d */
                if (lengthPL(pl) != 2 || nthPL(pl, 1)->ty != ATOM)
                    error("%p: usage: (use filename)", p);

                return mkUse(nthPL(pl, 1)->u.atom);
            }
            return mkExp(parseexp(p));
        }
    }
    assert(0);
    return NULL;
}
/* parse.c 512a */
static Namelist*
getnamelist(Name *f, Par *p, Parlist *pl)
{
    if (pl == NULL)
        return NULL;
    if (pl->hd->ty != ATOM)
        error("%p: list in formal parameter list of function %n", p, f);
    return mkNL(pl->hd->u.atom, getnamelist(f, p, pl->tl));
}
/* parse.c 513a */
static Explist*
parselist(Parlist *pl)
{
    Exp *e;

    if (pl == NULL)
        return NULL;

    e = parseexp(pl->hd);
    return mkEL(e, parselist(pl->tl));
}
/* parse.c 513c */
static Exp*
parseexp(Par *p)
{
    switch (p->ty) {
    case ATOM:
        /* parseexp [[atom]] and return the result 513d */
        {
            const char *s;
            char *t;
            long l;

            s = nametostr(p->u.atom);
            l = strtol(s, &t, 10);
            if (*t == '\0') /* the number is the whole string */
                return mkLiteral(l);
            else
                return mkVar(p->u.atom);
        }
    case LIST:
        /* parseexp [[list]] and return the result 514a */
        {
            Parlist *pl;
            Name *first;
            Explist *argl;

            pl = p->u.list;
            if (pl == NULL)
                error("%p: empty list in input", p);
            if (pl->hd->ty != ATOM)
                error("%p: first item of list not name", p);

            first = pl->hd->u.atom;
            argl  = parselist(pl->tl);
            if (first == strtoname("begin")) {
                /* parseexp [[begin]] and return the result 514b */
                return mkBegin(argl);
            } else if (first == strtoname("if")) {
                /* parseexp [[if]] and return the result 514c */
                if (lengthEL(argl) != 3)
                    error("%p: usage: (if cond true false)", p);
                return mkIfx(nthEL(argl, 0), nthEL(argl, 1), nthEL(argl, 2));
            } else if (first == strtoname("set")) {
                /* parseexp [[set]] and return the result 514e */
                if (lengthEL(argl) != 2)
                    error("%p: usage: (set var exp)", p);
                if (nthEL(argl, 0)->ty != VAR)
                    error("%p: set needs variable as first param", p);
                return mkSet(nthEL(argl, 0)->u.var, nthEL(argl, 1));
            } else if (first == strtoname("while")) {
                /* parseexp [[while]] and return the result 514d */
                if (lengthEL(argl) != 2)
                    error("%p: usage: (while cond body)", p);
                return mkWhilex(nthEL(argl, 0), nthEL(argl, 1));
            } else {
                /* parseexp function application and return the result 515a */
                return mkApply(first, argl);
            }
        }
    default:
        assert(0);
        return NULL;
    }
}
/* parse.c 515b */
struct Topreader {
    int doprompt;
    Reader *r;
    Parlist *pl;
};
/* parse.c 515c */
Toplevel*
readtop(Topreader *tr)
{
    Par *p;

    if (tr->pl == NULL) {
        tr->pl = readparlist(tr->r, 0, tr->doprompt);
        if (tr->pl == NULL) 
            return NULL;
    }

    p      = tr->pl->hd;
    tr->pl = tr->pl->tl;
    return parse(p);
}
/* parse.c 515d */
Topreader*
topreader(Reader *r, int doprompt)
{
    Topreader *tr;

    tr = malloc(sizeof(*tr));
    assert(tr != NULL);

    tr->r        = r;
    tr->doprompt = doprompt;
    tr->pl       = NULL;
    return tr;
}
