(* tuscheme.sml *)


(*****************************************************************)
(*                                                               *)
(*   ENVIRONMENTS                                                *)
(*                                                               *)
(*****************************************************************)

(* environments 186 *)
type name = string
type 'a env = (name * 'a) list
val emptyEnv = []

(* lookup and assignment of existing bindings *)
exception NotFound of name
fun find (name, []) = raise NotFound name
  | find (name, (n, v)::tail) = if name = n then v else find(name, tail)

(* adding new bindings *)
exception BindListLength
fun bind(name, v, rho) = (name, v) :: rho
fun bindList(n::vars, v::vals, rho) = bindList(vars, vals, bind(n, v, rho))
  | bindList([], [], rho) = rho
  | bindList _ = raise BindListLength
(* type declarations for consistency checking *)
val _ = op emptyEnv : 'a env
val _ = op find     : name * 'a env -> 'a
val _ = op bind     : name      * 'a      * 'a env -> 'a env
val _ = op bindList : name list * 'a list * 'a env -> 'a env
exception LeftAsExercise
exception TypeError of string


(*****************************************************************)
(*                                                               *)
(*   DEFINITION OF [[SEPARATE]]                                  *)
(*                                                               *)
(*****************************************************************)

(* definition of [[separate]] 190d *)
fun separate (zero, sep) =  (* useful for printing *)
  let fun s []     = zero
        | s [x]    = x
        | s (h::t) = h ^ sep ^ s t
  in  s
end


(*****************************************************************)
(*                                                               *)
(*   TYPES FOR {\TUSCHEME}                                       *)
(*                                                               *)
(*****************************************************************)

(* types for {\tuscheme} *)
datatype kind = TYPE                          (* kind of all types *)
              | ARROW of kind list * kind     (* kind of many constructors *)
(* types for {\tuscheme} *)
datatype tyex = TYCON  of name                (* type constructor *)
              | TYVAR  of name                (* type variable *)
              | CONAPP of tyex * tyex list    (* type-level application *)
              | FORALL of name list * tyex
(* types for {\tuscheme} *)
(* printing types for {\tuscheme} 585a *)
val spaceSep = separate("", " ")
fun typeString (TYCON c) = c
  | typeString (TYVAR a) = tyvarString a
  | typeString (CONAPP (TYCON "function", [CONAPP (TYCON "tuple", args), result]
                                                                            )) =
      "(function (" ^ spaceSep (map typeString args) ^ ") " ^ typeString result
                                                                           ^ ")"
  | typeString (CONAPP (tau, [])) = "(" ^ typeString tau ^ ")"
  | typeString (CONAPP (tau, l)) =
      "(" ^ typeString tau ^ " " ^ spaceSep (map typeString l) ^ ")"
  | typeString (FORALL (l, tau)) =
      "(forall (" ^ spaceSep (map tyvarString l) ^ ") " ^ typeString tau ^ ")"
and tyvarString a = "'" ^ a
(* type declarations for consistency checking *)
val _ = op typeString : tyex -> string
(* types for {\tuscheme} *)
val inttype  = TYCON "int"
val booltype = TYCON "bool"
val symtype  = TYCON "sym"
val unittype = TYCON "unit"
val tyvarA   = TYVAR "a"
fun tupletype l = CONAPP (TYCON "tuple", l)
fun listtype ty = CONAPP (TYCON "list",[ty])
fun funtype (args, result) = CONAPP (TYCON "function", [tupletype args, result])
(* type declarations for consistency checking *)
val _ = op inttype   : tyex
val _ = op booltype  : tyex
val _ = op symtype   : tyex
val _ = op unittype  : tyex
val _ = op tyvarA    : tyex
val _ = op tupletype : tyex list -> tyex
val _ = op listtype  : tyex -> tyex
val _ = op funtype   : tyex list * tyex -> tyex


(*****************************************************************)
(*                                                               *)
(*   LEXICAL ANALYSIS                                            *)
(*                                                               *)
(*****************************************************************)

(* lexical analysis ((mlscheme)) 571a *)
type srcloc  = string * int
fun srclocString (source, line) = source ^ ", line " ^ Int.toString line
(* lexical analysis ((mlscheme)) 571b *)
datatype par = NAME  of string
             | INT   of int
             | SHARP of char  (* #t or #f *)
             | LIST  of srcloc * par list
(* lexical analysis ((mlscheme)) 572 *)

exception Paren of char list * srcloc
   (* raise if found ) when reading input *)
exception PrematureEOF of string (* shows what we were reading ... *)
exception EOF                 (* raised by reader *)
fun read nextline curloc : string -> par list =
  let (* character-fiddling functions 573a *)
      fun isDelim c = List.exists (fn c' => c' = c) [#"(", #")", #" ", #"\t",
                                                                    #"\n", #";"]
      (* character-fiddling functions 573b *)
      fun stringReader []     = NONE
        | stringReader (h::t) = SOME (h, t)
      val strton = Int.scan StringCvt.DEC stringReader
      (* type declarations for consistency checking *)
      val _ = op strton : char list -> (int * char list) option
      (* character-fiddling functions 573c *)
      fun skip s =
        case StringCvt.skipWS stringReader s
          of #";" :: s => skip (StringCvt.dropl (fn c => c <> #"\n")
                                                                 stringReader s)
           | s => s
      (* character-fiddling functions 573d *)
      fun nonwhite s =
        case skip s
          of [] => nonwhite (explode (nextline()))
           | s  => s
      fun readmap f (result, unconsumed) = (f result, unconsumed)
      (* get is called only when s is guaranteed non-blank, non-comment, non-
                                                                        empty *)
      fun get (#"("::s)  = readmap (fn l => LIST l) (list (curloc(), s, []))
        | get (#")"::s)  = raise (Paren (s, curloc()))
        | get (#"'"::s)  = let val loc = curloc()
                               fun quote p = LIST (loc, [NAME "quote", p])
                           in  readmap quote (get (nonwhite s))
                           end
        | get s          = atom (s, [])
      and atom(c::s, r) = if isDelim c then (mkAtom(rev r), c::s) 
                          else atom(s, c::r)
        | atom([],   r) = (mkAtom(rev r), [])
      and mkAtom [#"#", #"f"] = SHARP #"f"
        | mkAtom [#"#", #"t"] = SHARP #"t"
        | mkAtom s = (case strton s of SOME (n, []) => INT n | _ => NAME (
                                                                     implode s))
      and list (loc, s, l) =
        let val next = nonwhite s
        in  let val (a, u) = get next
            in  list(loc, u, a::l)
            end handle (Paren (u, _)) => ((loc, rev l), skip u)
                     | EOF => raise PrematureEOF (implode next)
        end handle EOF => raise PrematureEOF (implode s)
      fun readline [] = []
        | readline s  = 
             let val(p, u) = get s
             in  p :: readline(skip u)
             end
(* type declarations for consistency checking *)
val _ = op read : (unit -> string) -> (unit -> srcloc) -> string -> par list
val _ = op get  : char list                      -> par                 * char
                                                                            list
val _ = op atom : char list * char list          -> par                 * char
                                                                            list
val _ = op list : srcloc * char list * par list  -> (srcloc * par list) * char
                                                                            list
  in  fn s => readline(skip (explode s)) 
  end


(*****************************************************************)
(*                                                               *)
(*   ABSTRACT SYNTAX AND VALUES FOR {\TUSCHEME}                  *)
(*                                                               *)
(*****************************************************************)

(* abstract syntax and values for {\tuscheme} *)
datatype       exp = LITERAL  of value
                   | VAR      of name
                   | SET      of name * exp
                   | IFX      of exp * exp * exp
                   | WHILEX   of exp * exp
                   | BEGIN    of exp list
                   | LETX     of let_kind * (name * exp) list * exp
                   | LAMBDA   of lambda_exp
                   | APPLY    of exp * exp list
                   | TYLAMBDA of name list * exp
                   | TYAPPLY  of exp * tyex list
and       let_kind = LET | LETSTAR
and          value = NIL
                   | BOOL      of bool   
                   | NUM       of int
                   | SYM       of name
                   | PAIR      of value * value
                   | CLOSURE   of lambda_value * value ref env
                   | PRIMITIVE of primitive
withtype primitive    = value list -> value (* raises RuntimeError *)
     and lambda_exp   = (name * tyex) list * exp
     and lambda_value = name          list * exp

exception RuntimeError of string
(* abstract syntax and values for {\tuscheme} *)
datatype toplevel = EXP    of exp
                  | DEFINE of name * tyex * lambda_exp
                  | VAL    of name * exp
                  | VALREC of name * tyex * exp
                  | USE    of name


(*****************************************************************)
(*                                                               *)
(*   VALUES                                                      *)
(*                                                               *)
(*****************************************************************)

(* values ((mlscheme)) 188b *)



fun embedList []     = NIL
  | embedList (h::t) = PAIR(h, embedList t)
fun embedPredicate f = fn x => BOOL (f x)
fun bool (BOOL b) = b
  | bool _        = true
(* type declarations for consistency checking *)
val _ = op embedList      : value list -> value
val _ = op embedPredicate : ('a -> bool) -> ('a -> value)
val _ = op bool           : value -> bool
(* values ((mlscheme)) 189a *)
fun valueString (NIL)    = "()"
  | valueString (BOOL b) = if b then "#t" else "#f"
  | valueString (NUM n)  = String.map (fn #"~" => #"-" | c => c) (Int.toString n
                                                                               )
  | valueString (SYM v)  = v
  | valueString (PAIR(car, cdr))  = 
      let fun tail (PAIR(car, cdr)) = " " ^ valueString car ^ tail cdr
            | tail NIL = ")"
            | tail v = " . " ^ valueString v ^ ")"
      in  "(" ^ valueString car ^ tail cdr
      end
  | valueString (CLOSURE   _) = "<procedure>"
  | valueString (PRIMITIVE _) = "<procedure>"
(* type declarations for consistency checking *)
val _ = op valueString : value -> string


(*****************************************************************)
(*                                                               *)
(*   TYPE CHECKING FOR {\TUSCHEME}                               *)
(*                                                               *)
(*****************************************************************)

(* type checking for {\tuscheme} *)
fun kindof(tau, Delta) =
  let (* internal function [[kind]] *)
      fun kind (TYVAR a) =
            (find("'" ^ a, Delta)
             handle NotFound _ => raise TypeError ("unknown type variable '" ^ a
                                                                              ))
      (* internal function [[kind]] *)
        | kind (TYCON c) =
            (find(c, Delta)
             handle NotFound _ => raise TypeError ("unknown type constructor " ^
                                                                             c))
      (* internal function [[kind]] *)
        | kind (CONAPP (TYCON "tuple", actuals)) =
            if List.all (fn tau => kind tau = TYPE) actuals then
                TYPE
            else
                raise TypeError "tuple formed from non-types"
      (* internal function [[kind]] *)
        | kind (CONAPP (tau, actuals)) =
            (case kind tau
               of ARROW (formals, result) =>
                    if formals = map kind actuals then
                        result
                    else
                        raise TypeError ("type constructor " ^ typeString tau ^
                                         " applied to the wrong arguments")
                | TYPE =>
                    raise TypeError ("tried to apply type " ^ typeString tau ^
                                     " as type constructor"))
      (* internal function [[kind]] *)
        | kind (FORALL (alphas, tau)) =
            let val Delta' =
                  foldl (fn (a, Delta) => bind("'" ^ a, TYPE, Delta)) Delta
                                                                          alphas
            in  case kindof(tau, Delta')
                  of TYPE    => TYPE
                   | ARROW _ => raise TypeError
                                      "quantifed a non-nullary type constructor"
            end
(* type declarations for consistency checking *)
val _ = op kindof : tyex * kind env -> kind
val _ = op kind   : tyex            -> kind
  in  kind tau
  end
(* type checking for {\tuscheme} *)
fun asType (ty, Delta) =
  case kindof(ty, Delta)
    of TYPE    => ty
     | ARROW _ => raise TypeError ("used type constructor `" ^ typeString ty ^
                                   "' as a type")
(* type declarations for consistency checking *)
val _ = op asType : tyex * kind env -> tyex
(* type checking for {\tuscheme} *)
fun tysubst (tau, varenv) =
  let fun subst (TYVAR a) = (find(a, varenv) handle NotFound _ => TYVAR a)
        | subst (TYCON c) = (TYCON c)
        | subst (CONAPP (tau, taus)) = CONAPP (subst tau, map subst taus)
        | subst (FORALL (alphas, tau)) =
            FORALL (alphas, tysubst(tau, bindList(alphas, map TYVAR alphas,
                                                                       varenv)))
(* type declarations for consistency checking *)
val _ = op tysubst : tyex * tyex env -> tyex
val _ = op subst   : tyex            -> tyex
  in  subst tau
  end
(* type checking for {\tuscheme} *)
fun instantiate (FORALL (formals, tau), actuals, Delta) =
      (case List.find (fn t => kindof(t, Delta) <> TYPE) actuals
         of SOME t => raise TypeError ("instantiated at type constructor `" ^
                                       typeString t ^ "', which is not a type")
          | NONE =>
              (tysubst(tau, bindList (formals, actuals, emptyEnv))
               handle BindListLength =>
                 raise TypeError
                   "instantiated polymorphic term at wrong number of arguments")
                                                                               )
  | instantiate _ = raise TypeError
                             "tried to instantiate term of non-polymorphic type"
(* type declarations for consistency checking *)
val _ = op instantiate : tyex * tyex list * kind env -> tyex
val _ = List.find : ('a -> bool) -> 'a list -> 'a option
(* type checking for {\tuscheme} *)
fun eqType (TYCON c, TYCON c') = c = c'
  | eqType (TYVAR a, TYVAR a') = a = a'
  | eqType (CONAPP (tau, taus), CONAPP (tau', taus')) =
      eqType (tau, tau') andalso eqTypes (taus, taus')
  | eqType (FORALL (alphas, tau), FORALL (alphas', tau')) =
      (eqType (tau, tysubst (tau', bindList (alphas', map TYVAR alphas, emptyEnv
                                                                             )))
       handle BindListLength => false)
  | eqType _ = false
and eqTypes (t::taus, t'::taus') = eqType(t, t') andalso eqTypes(taus, taus')
  | eqTypes ([], []) = true
  | eqTypes _ = false
(* type declarations for consistency checking *)
val _ = op eqType  : tyex      * tyex      -> bool
val _ = op eqTypes : tyex list * tyex list -> bool
(* type checking for {\tuscheme} *)
fun appearsUnprotectedIn (x, e) = 
  let fun evaluatesX (LITERAL n) = false
        | evaluatesX (VAR v) = v = x
        | evaluatesX (SET (_, e)) = evaluatesX e
        | evaluatesX (WHILEX (e1, e2)) = evaluatesX e1 orelse evaluatesX e2
        | evaluatesX (APPLY (f, actuals)) =
            evaluatesX f orelse List.exists evaluatesX actuals
        | evaluatesX (LETX (LETSTAR, [], body)) = evaluatesX body
        | evaluatesX (LETX (LETSTAR, (x', e') :: bs, body)) =
            evaluatesX e' orelse
            (x <> x' andalso evaluatesX (LETX (LETSTAR, bs, body)))
        | evaluatesX (LETX (LET, bs, body)) = 
            List.exists (fn (_, e) => evaluatesX e) bs orelse
            (not (List.exists (fn (x', _) => x' = x) bs) andalso evaluatesX body
                                                                               )
        | evaluatesX (IFX (e1, e2, e3)) =
            evaluatesX e1 orelse evaluatesX e2 orelse evaluatesX e3
        | evaluatesX (BEGIN es) = List.exists evaluatesX es
        | evaluatesX (LAMBDA (formals, body)) = false
        | evaluatesX (TYAPPLY (e, args)) = evaluatesX e
        | evaluatesX (TYLAMBDA (alphas, e)) = evaluatesX e
  in  evaluatesX e
  end
(* type checking for {\tuscheme} *)
exception LeftAsExercise
fun topty _ = raise LeftAsExercise
(* type declarations for consistency checking *)
val _ = op topty : toplevel * tyex env * kind env -> tyex env * string


(*****************************************************************)
(*                                                               *)
(*   PARSING FOR {\TUSCHEME}                                     *)
(*                                                               *)
(*****************************************************************)

(* parsing for {\tuscheme} 585b *)
exception SyntaxError of string
fun syntaxErrorAt (source, line) msg =
  raise SyntaxError (source ^ ", line " ^ Int.toString line ^ ": " ^ msg)
fun nodups nameOf msg loc l =
  let fun dup [] = l
        | dup (h::t) = if List.exists (fn x => nameOf x = nameOf h) t then
                         syntaxErrorAt loc (msg h)
                       else
                         dup t
  in  dup l
  end
(* type declarations for consistency checking *)
val _ = op nodups : ('a -> name) -> ('a -> string) -> srcloc -> 'a list -> 'a
                                                                            list
(* parsing for {\tuscheme} 586 *)
val id  = fn n => n
val fst = fn (n, _) => n
fun defineDups loc f =
  nodups fst (fn x => "formal parameter " ^ fst x ^
                      " appears twice in definition of function " ^ f) loc
fun lambdaDups loc =
  nodups fst (fn x => "formal parameter " ^ fst x ^ " appears twice in lambda")
                                                                             loc
fun tylambdaDups loc =
  nodups id (fn x => "formal parameter " ^ x ^ " appears twice in type-lambda")
                                                                             loc
fun forallDups loc =
  nodups id (fn x => "quantified variable " ^ x ^ " appears twice in forall")
                                                                             loc
fun letDups loc (LETSTAR, bindings) = bindings
  | letDups loc (LET,     bindings) =
      nodups fst (fn x => "name " ^ fst x ^ " appears twice in let") loc
                                                                        bindings
(* parsing for {\tuscheme} 587 *)
fun parseType _ (LIST (loc, NAME "forall"   :: args)) = parseForall   loc args
  | parseType _ (LIST (loc, NAME "function" :: args)) = parseFunction loc args
  | parseType _ (LIST (loc, [NAME "quote", NAME a]))  = TYVAR a
  | parseType _ (LIST (loc, NAME (bad as "quote")  :: args)) =
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "if")     :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "while")  :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "set")    :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "begin")  :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "lambda") :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "type-lambda") :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "let")    :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "let*")   :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, NAME (bad as "@")      :: args)) = 
      syntaxErrorAt loc ("looking for type but found `" ^ bad ^ "'")
  | parseType _ (LIST (loc, tau :: taus)) =
      CONAPP (parseType loc tau, map (parseType loc) taus)
  | parseType _ (LIST (loc, [])) = syntaxErrorAt loc "empty type ()"
  | parseType _ (NAME a) = TYCON a
  | parseType loc (INT n)   = syntaxErrorAt loc "expected type; found integer"
  | parseType loc (SHARP c) = syntaxErrorAt loc
                                          "expected type; found Boolean literal"
and parseForall loc [LIST (floc, formals), tau] =
      FORALL (forallDups floc (map (untyvar floc) formals), parseType loc tau)
  | parseForall loc _ = syntaxErrorAt loc "expected (forall (names) type)"
and parseFunction loc [LIST (aloc, args), tau] =
      funtype (map (parseType aloc) args, parseType loc tau)
  | parseFunction loc _ = syntaxErrorAt loc "expected (function (types) type)"
and untyvar _ (LIST (_, [NAME "quote", NAME n])) = n
  | untyvar loc _ = syntaxErrorAt loc "type parameters must be type variables"
(* parsing for {\tuscheme} 588 *)
fun parse (NAME s)  = VAR s
  | parse (INT n)   = LITERAL (NUM n)
  | parse (SHARP c) = LITERAL (BOOL (c = #"t"))
  | parse (LIST (loc, NAME "quote"  :: args)) = parseQuote loc args
  | parse (LIST (loc, NAME "if"     :: args)) = parseIf    loc args
  | parse (LIST (loc, NAME "while"  :: args)) = parseWhile loc args
  | parse (LIST (loc, NAME "set"    :: args)) = parseSet   loc args
  | parse (LIST (loc, NAME "begin"  :: args)) = BEGIN (map parse args)
  | parse (LIST (loc, NAME "lambda" :: args)) = parseLambda loc args
  | parse (LIST (loc, NAME "type-lambda" :: args)) = parseTypeLambda loc args
  | parse (LIST (loc, NAME "let"    :: args)) = parseLet loc LET     args
  | parse (LIST (loc, NAME "let*"   :: args)) = parseLet loc LETSTAR args
  | parse (LIST (loc, NAME "@"      :: args)) = parseTyapp loc args
  | parse (LIST (loc, NAME "forall" :: _   )) =
       syntaxErrorAt loc "looking for expression but found `forall'"
  | parse (LIST (loc, NAME "function" :: _   )) =
       syntaxErrorAt loc "looking for expression but found `function"
  | parse (LIST (loc, h::t))  = APPLY (parse h, map parse t)
  | parse (LIST (loc, [])) = syntaxErrorAt loc "empty application ()"
(* parsing for {\tuscheme} 589a *)
and parseQuote loc [e] = LITERAL (quote loc e)
  | parseQuote loc _ = syntaxErrorAt loc "expected (quote e)"
and parseIf _ [e1, e2, e3] = IFX (parse e1, parse e2, parse e3)
  | parseIf loc _ = syntaxErrorAt loc "expected (if e1 e2 e3)"
and parseWhile _ [e1, e2] = WHILEX (parse e1, parse e2)
  | parseWhile loc _ = syntaxErrorAt loc "expected (while e1 e2)"
and parseSet _ [NAME v, e] = SET (v, parse e)
  | parseSet loc _ = syntaxErrorAt loc "expected (set v e)"
and parseTyapp loc (exp :: args) = TYAPPLY (parse exp, map (parseType loc) args)
  | parseTyapp loc [] = syntaxErrorAt loc "empty type application (@)"
and parseTypeLambda _ [LIST (loc, formals), body] =
      TYLAMBDA (tylambdaDups loc (map (untyvar loc) formals), parse body)
  | parseTypeLambda loc _ =
      syntaxErrorAt loc "expected (lambda (names) body)"
and parseLambda _ [LIST (loc, formals), body] =
      LAMBDA (lambdaDups loc (map (unformal loc) formals), parse body)
  | parseLambda loc _ =
      syntaxErrorAt loc "expected (lambda (formals) body)"
and parseLet _ kind [LIST (loc, bindings), body] = 
      LETX(kind, letDups loc (kind, map (unBinding loc) bindings), parse body)
  | parseLet loc _ _ = syntaxErrorAt loc "expected (let (bindings) body)"
and unformal _ (LIST (loc, [tau, NAME n])) = (n, parseType loc tau)
  | unformal _ (LIST (loc,_)) = syntaxErrorAt loc
                                          "formal parameter must be (type name)"
  | unformal loc _ = syntaxErrorAt loc "formal parameter must be (type name)"
and unBinding _ (LIST (_, [NAME n, e])) = (n, parse e)
  | unBinding _ (LIST (loc, _)) = syntaxErrorAt loc "expected (v e) in bindings"
  | unBinding loc _ = syntaxErrorAt loc "expected (v e) in bindings"
and quote _ (INT   n) = NUM n
  | quote _ (SHARP c) = BOOL (c = #"t")
  | quote loc (NAME  ".") =
      syntaxErrorAt loc
                      "this interpreter cannot handle . in quoted S-expressions"
  | quote _ (NAME  n) = SYM n
  | quote _ (LIST (loc, l)) = embedList (map (quote loc) l)
(* parsing for {\tuscheme} 589b *)
fun toplevel (LIST (loc, NAME "define"  :: args)) = parseDefine loc args
  | toplevel (LIST (loc, NAME "val"     :: args)) = parseVal    loc args
  | toplevel (LIST (loc, NAME "val-rec" :: args)) = parseValRec loc args
  | toplevel (LIST (loc, NAME "global"  :: args)) = parseVal    loc args (*
                                                                       legacy *)
  | toplevel (LIST (loc, NAME "use"     :: args)) = parseUse    loc args
  | toplevel e = EXP (parse e)
and parseDefine loc [ty, NAME name, LIST (floc, formals), body] =
      DEFINE (name, parseType loc ty,
              (defineDups floc name (map (unformal floc) formals), parse body))
  | parseDefine loc _ = syntaxErrorAt loc "expected (define type f (args) body)"
and parseVal _ [NAME x, e] = VAL (x, parse e)
  | parseVal loc _ = syntaxErrorAt loc "expected (val x e)"
and parseValRec loc [tau, NAME x, e] = VALREC (x, parseType loc tau, parse e)
  | parseValRec loc _ = syntaxErrorAt loc "expected (val-rec type x e)"
and parseUse _ [NAME x] = USE x
  | parseUse loc _ = syntaxErrorAt loc "expected (use filename)"


(*****************************************************************)
(*                                                               *)
(*   READERS                                                     *)
(*                                                               *)
(*****************************************************************)

(* readers 576b *)
type reader = { firstline : unit -> string
              , nextline  : unit -> string
              , curloc    : unit -> string * int
              }
(* readers 576c *)
fun promptReader (firstPrompt, nextPrompt, reader) =
  let val { firstline = first, nextline = next, curloc = cur } = reader
      fun promptIn get prompt () = ( TextIO.output(TextIO.stdOut, prompt)
                                   ; TextIO.flushOut(TextIO.stdOut)
                                   ; get ()
                                   )
  in  { firstline = promptIn first firstPrompt
      , nextline  = promptIn next  nextPrompt
      , curloc    = cur
      }
  end
(* type declarations for consistency checking *)
val _ = op promptReader : string * string * reader -> reader
(* readers 577a *)
fun filereader (filename, fd)=
  let val lineNum = ref 0
      fun get () =
        let val line = TextIO.inputLine fd
        in  if size line = 0 then raise EOF else line before lineNum := !lineNum
                                                                             + 1
        end
  in  { firstline = get, nextline = get, curloc = (fn () => (filename, !lineNum)
                                                                             ) }
  end

fun stringsreader (what, l) = 
  let val buffer  = ref l
      val lineNum = ref 0
      fun get ()  = case !buffer
                      of [] => raise EOF
                       | h :: t => h before buffer := t before lineNum := !
                                                                     lineNum + 1
  in  { firstline = get, nextline = get, curloc = (fn () => (what, !lineNum)) }
  end              
(* type declarations for consistency checking *)
val _ = op filereader    : string * TextIO.instream -> reader
val _ = op stringsreader : string * string list     -> reader
(* readers 577b *)
type topreader = { buffer : toplevel list ref, reader : reader }
fun readtop (r as { buffer, reader = { nextline, firstline, curloc } }) =
  case !buffer
    of h::t => h before buffer := t
     | []   => ( buffer := map toplevel (read nextline curloc (firstline()))
               ; readtop r
               )
(* readers 577c *)
fun topreader (reader, prompt) = 
  let val reader = if prompt then promptReader ("-> ", "   ", reader) else
                                                                          reader
  in  { buffer = ref [], reader = reader }
  end
(* readers 578 *)
fun echoTag f x =
  let val line = f x
      val _ = if (String.substring (line, 0, 2) = ";#" handle _ => false) then
                print line
              else
                ()
  in  line
  end

fun echoReader { nextline=n, firstline=f, curloc=c } =
     { nextline=echoTag n, firstline=echoTag f, curloc=c } 

fun echoTop { buffer = b, reader = r } = { buffer = b, reader = echoReader r }
val topreader = fn args => echoTop (topreader args)
(* type declarations for consistency checking *)
val _ = op echoReader : reader -> reader 


(*****************************************************************)
(*                                                               *)
(*   IMPLEMENTATION OF [[USE]]                                   *)
(*                                                               *)
(*****************************************************************)

(* implementation of [[use]] 193c *)
fun use readEvalPrint filename rho =
      let val fd = TextIO.openIn filename
          val treader = topreader (filereader (filename, fd), false)
          fun writeln s = app print [s, "\n"]
      in  readEvalPrint (treader, writeln, writeln) rho
          before TextIO.closeIn fd
      end 


(*****************************************************************)
(*                                                               *)
(*   CHECKING AND EVALUATION FOR {\TUSCHEME}                     *)
(*                                                               *)
(*****************************************************************)

(* checking and evaluation for {\tuscheme} *)
exception BugInTypeChecking of string
(* evaluation for {\tuscheme} 590a *)
fun eval(e, rho) =
  let fun ev(LITERAL n) = n
        (* alternatives for [[ev]] for [[TYAPPLY]] and [[TYLAMBDA]] *)
        | ev(TYAPPLY  (e, _)) = ev e
        | ev(TYLAMBDA (_, e)) = ev e
        (* more alternatives for [[ev]] for {\tuscheme} 590b *)
        | ev(VAR v) = !(find(v, rho))
        | ev(SET (n, e)) = 
            let val v = ev e
            in  find (n, rho) := v;
                v
            end
        (* more alternatives for [[ev]] for {\tuscheme} 590c *)
        | ev(IFX (e1, e2, e3)) = ev (if bool (ev e1) then e2 else e3)
        | ev(WHILEX (guard, body)) = 
            if bool (ev guard) then 
              (ev body; ev(WHILEX (guard, body)))
            else
              NIL
        | ev(BEGIN es) =
            let fun b(e::es, lastval) = b(es, ev e)
                  | b(   [], lastval) = lastval
            in  b(es, BOOL false)
            end
        (* more alternatives for [[ev]] for {\tuscheme} 590d *)
        | ev(LAMBDA (args, body)) = CLOSURE((map (fn (n, ty) => n) args, body),
                                                                            rho)
        (* more alternatives for [[ev]] for {\tuscheme} 590e *)
        | ev(APPLY (f, args))  = 
               (case ev f
                  of PRIMITIVE prim => prim (map ev args)
                   | CLOSURE clo => (* apply closure [[clo]] to [[args]] ((
                                                              mlscheme)) 190c *)
                                    let val ((formals, body), savedrho) = clo
                                        val actuals = map ev args
                                    in  eval(body, bindList(formals, map ref
                                                             actuals, savedrho))
                                        handle BindListLength => 
                                            raise RuntimeError (
                                      "Wrong number of arguments to closure; " ^
                                                                "expected (" ^
                                                separate("", " ") formals ^ ")")
                                    end
                   | v => raise BugInTypeChecking "applied non-function"
               )
        (* more alternatives for [[ev]] for {\tuscheme} 591a *)
        | ev(LETX (LET, bs, body)) =
            let val (names, values) = ListPair.unzip bs
            in  eval (body, bindList(names, map (ref o ev) values, rho))
            end
        | ev(LETX (LETSTAR, bs, body)) =
            let fun step ((n, e), rho) = bind(n, ref (eval(e, rho)), rho)
            in  eval (body, foldl step rho bs)
            end
(* type declarations for consistency checking *)
val _ = op eval : exp * value ref env -> value
val _ = op ev   : exp                 -> value
  in  ev e
  end
(* evaluation for {\tuscheme} 591b *)
fun topeval (t, rho) =
  case t
    of USE filename => raise RuntimeError
                                       "internal error -- `use' reached topeval"
     | EXP e        => (rho, valueString (eval(e, rho)))
     | DEFINE (name, tau, lambda) => topeval (VALREC (name, tau, LAMBDA lambda),
                                                                            rho)
     | VAL    (name, e)      =>
          let val v   = eval(e, rho)
              val rho = bind(name, ref v, rho)
          in  (rho, showVal name v)
          end
     | VALREC (name, tau, e) => 
          let val rho = bind(name, ref NIL, rho)
              val v   = eval(e, rho)
          in  find(name, rho) := v;
              (rho, showVal name v)
          end
(* evaluation for {\tuscheme} 591c *)
and showVal name v =
      case v
        of CLOSURE   _ => name
         | PRIMITIVE _ => name
         | _ => valueString v
(* type declarations for consistency checking *)
val _ = op topeval : toplevel * value ref env -> value ref env * string
(* evaluation for {\tuscheme} 592a *)
fun binaryOp f = (fn [a, b] => f(a, b) | _ => raise BugInTypeChecking "arity 2")
fun unaryOp  f = (fn [a]    => f a     | _ => raise BugInTypeChecking "arity 1")
(* type declarations for consistency checking *)
val _ = op unaryOp  : (value         -> value) -> (value list -> value)
val _ = op binaryOp : (value * value -> value) -> (value list -> value)
(* evaluation for {\tuscheme} 592b *)
fun arithOp f =
      binaryOp (fn (NUM n1, NUM n2) => NUM (f(n1, n2)) 
                 | _ => raise BugInTypeChecking "arithmetic on non-numbers")
val arithtype = funtype ([inttype, inttype], inttype)
(* type declarations for consistency checking *)
val _ = op arithOp   : (int * int -> int) -> (value list -> value)
val _ = op arithtype : tyex
(* evaluation for {\tuscheme} 592d *)
fun embedPredicate f args = BOOL (f args)
fun comparison f = binaryOp (embedPredicate f)
fun intcompare f = 
      comparison (fn (NUM n1, NUM n2) => f(n1, n2)
                   | _ => raise BugInTypeChecking "comparing non-numbers")
val comptype = funtype ([inttype, inttype], booltype)
(* type declarations for consistency checking *)
val _ = op comparison : (value * value -> bool) -> (value list -> value)
val _ = op intcompare : (int   * int   -> bool) -> (value list -> value)
val _ = op comptype   : tyex
type env_bundle = kind env * tyex env * value ref env
fun topCheckEval (t, envs as (delta, gamma, rho), echo) =
  case t
    of USE filename => use readCheckEvalPrint filename envs
     | _ =>
        let val (gamma, tystring)  = topty   (t, gamma, delta)
            val (rho,   valstring) = topeval (t, rho)
            val _ = if size valstring > 0 then echo (valstring ^ " : " ^
                                                                       tystring)
                    else ()
(* type declarations for consistency checking *)
val _ = op topCheckEval : toplevel * env_bundle * (string->unit) -> env_bundle
        in  (delta, gamma, rho)
        end
(* checking and evaluation for {\tuscheme} *)
and readCheckEvalPrint (reader, echo, errmsg) envs =
  let fun loop envs =
        let fun continue msg = (errmsg msg; loop envs)
            fun finish() = envs
        in  loop (topCheckEval (readtop reader, envs, echo))
            handle EOF => finish()
                (* more read-eval-print handlers *)
                | TypeError         msg => continue ("type error: " ^ msg)
                | BugInTypeChecking msg => continue ("bug in type checking: " ^
                                                                            msg)
                (* more read-eval-print handlers 194b *)
                | IO.Io {name, ...} => continue ("I/O error: " ^ name)
                | Paren (_, loc)    => continue ("syntax error: extra ) in " ^
                                                               srclocString loc)
                | PrematureEOF s    => (errmsg ("Missing )? Got EOF reading (" ^
                                                                   s); finish())
                | SyntaxError msg   => continue ("syntax error: " ^ msg)
                (* more read-eval-print handlers 194c *)
                | Div               => continue "Division by zero"
                | Overflow          => continue "Arithmetic overflow"
                | RuntimeError msg  => continue ("run-time error: " ^ msg)
                | NotFound n        => continue ("variable " ^ n ^ " not found")
        end
  in  loop envs
  end
(* type declarations for consistency checking *)
val _ = op readCheckEvalPrint : topreader * (string->unit) * (string->unit) ->
                                                        env_bundle -> env_bundle


(*****************************************************************)
(*                                                               *)
(*   INITIALIZATION FOR {\TUSCHEME}                              *)
(*                                                               *)
(*****************************************************************)

(* initialization for {\tuscheme} *)
val initialEnvs =
  let fun addPrim ((name, prim, funty), (tfuns, vfuns)) = 
        ( bind (name, funty, tfuns)
        , bind (name, ref (PRIMITIVE prim), vfuns)
        )
      val (tfuns, vfuns)  = foldl addPrim (emptyEnv, emptyEnv)
                            ((* primops for {\tuscheme}\ [[::]] 592c *)
                             ("+", arithOp op +,   arithtype) :: 
                             ("-", arithOp op -,   arithtype) :: 
                             ("*", arithOp op *,   arithtype) :: 
                             ("/", arithOp op div, arithtype) ::
                             (* primops for {\tuscheme}\ [[::]] 592e *)
                             ("<", intcompare op <, comptype) :: 
                             (">", intcompare op >, comptype) ::
                             ("=", comparison (fn (NIL,     NIL    ) => true
                                                | (NUM  n1, NUM  n2) => n1 = n2
                                                | (SYM  v1, SYM  v2) => v1 = v2
                                                | (BOOL b1, BOOL b2) => b1 = b2
                                                |  _                 => false)
                                 , FORALL (["a"], funtype ([tyvarA, tyvarA],
                                                                  booltype))) ::
                             (* primops for {\tuscheme}\ [[::]] 593a *)
                             ("null?", unaryOp (embedPredicate (fn (NIL   ) =>
                                                             true | _ => false))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                                  booltype))) ::
                             ("cons", binaryOp (fn (a, b) => PAIR(a, b))
                                 , FORALL (["a"], funtype ([tyvarA, listtype
                                                  tyvarA], listtype tyvarA))) ::
                             ("car",  unaryOp  (fn (PAIR(car, _)) => car 
                                                 | v => raise RuntimeError
                                                                (
                                    "car applied to non-list " ^ valueString v))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                                    tyvarA))) ::
                             ("cdr",  unaryOp  (fn (PAIR(_, cdr)) => cdr 
                                                 | v => raise RuntimeError
                                                                (
                                    "cdr applied to non-list " ^ valueString v))
                                 , FORALL (["a"], funtype ([listtype tyvarA],
                                                           listtype tyvarA))) ::
                             (* primops for {\tuscheme}\ [[::]] 593b *)
                             ("print", unaryOp (fn x => (print (valueString x^
                                                                     "\n"); x)),
                                 FORALL (["a"], funtype ([tyvarA], unittype)))
                                                                         :: nil)
      fun addKind ((name, kind), kinds) = bind(name, kind, kinds)
      val kinds   = foldl addKind emptyEnv
                    ((* primitive type constructors for {\tuscheme}\ [[::]] *)
                     ("int",      TYPE) ::
                     ("bool",     TYPE) ::
                     ("sym",      TYPE) ::
                     ("unit",     TYPE) ::
                     ("list",     ARROW ([TYPE],       TYPE)) ::
                     ("function", ARROW ([TYPE, TYPE], TYPE)) :: nil)
      val envs    = (kinds, tfuns, vfuns)
      val basis   = (* ML representation of initial basis *)
                      [ "; tuscheme.nw 573"
                      ,
   "(val list1 (type-lambda ('a) (lambda (('a x)) ((@ cons 'a) x (@ '() 'a)))))"
                      , "(val list2 (type-lambda ('a) (lambda (('a x) ('a y))"
                      ,
            "                               ((@ cons 'a) x ((@ list1 'a) y)))))"
                      ,
                   "(val list3 (type-lambda ('a) (lambda (('a x) ('a y) ('a z))"
                      ,
          "                               ((@ cons 'a) x ((@ list2 'a) y z)))))"
                      , "; tuscheme.nw 593"
                      , "(val o (type-lambda ('a 'b 'c) "
                      , "  (lambda (((function ('b) 'c) f)"
                      , "           ((function ('a) 'b) g))"
                      , "     (lambda (('a x)) (f (g x))))))"
                      , ""
                      , "(val curry (type-lambda ('a 'b 'c)"
                      , "   (lambda (((function ('a 'b) 'c) f)) "
                      , "      (lambda (('a x)) (lambda (('b y)) (f x y))))))"
                      , ""
                      , "(val uncurry (type-lambda ('a 'b 'c)"
                      , "   (lambda (((function ('a) (function ('b) 'c)) f))"
                      , "      (lambda (('a x) ('b y)) ((f x) y)))))"
                      , "; tuscheme.nw 628"
                      , "(val-rec"
                      ,
                  "  (forall ('a) (function ((list 'a)) int))  ; type of length"
                      , "  length                                    ; name"
                      ,
    "  (type-lambda ('a)                         ; value : polymorphic function"
                      , "     (lambda (((list 'a) l))"
                      , "       (if ((@ null? 'a) l) 0"
                      , "          (+ 1 ((@ length 'a) ((@ cdr 'a) l)))))))"
                      , "; tuscheme.nw 639"
                      ,
      "(val-rec (forall ('a) (function ((list 'a) (list 'a)) (list 'a))) revapp"
                      , "  (type-lambda ('a)"
                      , "     (lambda (((list 'a) l1)  ((list 'a) l2))"
                      , "        (if ((@ null? 'a) l1)"
                      , "        l2"
                      ,
  "        ((@ revapp 'a) ((@ cdr 'a) l1) ((@ cons 'a) ((@ car 'a) l1) l2))))))"
                      , "; tuschemea.nw 387"
                      , "(val caar"
                      , "   (type-lambda ('a)"
                      , "      (lambda (((list (list 'a)) l))"
                      , "          ((@ car 'a) ((@ car (list 'a)) l)))))"
                      , "(val cadr"
                      , "   (type-lambda ('a)"
                      , "      (lambda (((list (list 'a)) l))"
                      , "          ((@ car (list 'a)) ((@ cdr (list 'a)) l)))))"
                      , "; tuschemea.nw 398"
                      , "(define bool and ((bool b) (bool c)) (if b  c  b))"
                      , "(define bool or  ((bool b) (bool c)) (if b  b  c))"
                      , "(define bool not ((bool b))          (if b #f #t))"
                      , "; tuschemea.nw 404"
                      ,
      "(val-rec (forall ('a) (function ((list 'a) (list 'a)) (list 'a))) append"
                      , "  (type-lambda ('a)"
                      , "     (lambda (((list 'a) l1)  ((list 'a) l2))"
                      , "       (if ((@ null? 'a) l1)"
                      , "         l2"
                      ,
 "         ((@ cons 'a) ((@ car 'a) l1) ((@ append 'a) ((@ cdr 'a) l1) l2))))))"
                      , "; tuschemea.nw 446"
                      ,
"(val-rec (forall ('a) (function ((function ('a) bool) (list 'a)) (list 'a))) filter"
                      , "   (type-lambda ('a)"
                      ,
                      "      (lambda (((function ('a) bool) p?)  ((list 'a) l))"
                      , "         (if ((@ null? 'a) l)"
                      , "             (@ '() 'a)"
                      , "             (if (p? ((@ car 'a) l))"
                      ,
"                 ((@ cons 'a) ((@ car 'a) l) ((@ filter 'a) p? ((@ cdr 'a) l)))"
                      ,
                       "                 ((@ filter 'a) p? ((@ cdr 'a) l)))))))"
                      , "; missing exists?"
                      , "; missing all?"
                      , "; tuschemea.nw 459"
                      ,
"(val-rec (forall ('a 'b) (function ((function ('a) 'b) (list 'a)) (list 'b))) map"
                      , "   (type-lambda ('a 'b)"
                      , "      (lambda (((function ('a) 'b) f)  ((list 'a) l))"
                      , "         (if ((@ null? 'a) l)"
                      , "             (@ '() 'b)"
                      ,
"             ((@ cons 'b) (f ((@ car 'a) l)) ((@ map 'a 'b) f ((@ cdr 'a) l)))))))"
                      , "; tuschemea.nw 479"
                      , "(define bool <= ((int x) (int y)) (not (> x y)))"
                      , "(define bool >= ((int x) (int y)) (not (< x y)))"
                      ,
     "(val != (type-lambda ('a) (lambda (('a x) ('a y)) (not ((@ = 'a) x y)))))"
                      , "; tuschemea.nw 486"
                      , "(define int max ((int x) (int y)) (if (> x y) x y))"
                      , "(define int min ((int x) (int y)) (if (< x y) x y))"
                      , ""
                      , "(define int mod ((int m) (int n)) (- m (* n (/ m n))))"
                      ,
   "(define int gcd ((int m) (int n)) (if ((@ = int) n 0) m (gcd n (mod m n))))"
                      ,
                      "(define int lcm ((int m) (int n)) (* m (/ n (gcd m n))))"
                      ]
(* type declarations for consistency checking *)
val _ = op kinds : kind      env
val _ = op tfuns : tyex      env
val _ = op vfuns : value ref env
val _ = op envs  : env_bundle
      val treader = topreader (stringsreader ("initial basis", basis), false)
  in  readCheckEvalPrint (treader, fn _ => (), fn _ => ()) envs
  end
(* initialization for {\tuscheme} *)
fun runInterpreter noisy = 
  let fun writeln s = app print [s, "\n"]
      val treader = topreader (filereader ("standard input", TextIO.stdIn),
                                                                          noisy)
  in  ignore (readCheckEvalPrint (treader, writeln, writeln) initialEnvs)
  end 


(*****************************************************************)
(*                                                               *)
(*   COMMAND LINE                                                *)
(*                                                               *)
(*****************************************************************)

(* command line *)
fun main ["-q"] = runInterpreter false
  | main []     = runInterpreter true
  | main _      =
      TextIO.output(TextIO.stdErr, "Usage: " ^ CommandLine.name() ^ " [-q]\n")
val _ = main (CommandLine.arguments())
