/* copy.c 166a */
#include "all.h"

/* private declarations for copying collection 166b */
static Value *fromspace, *tospace;           /* used only at GC time */
static unsigned semispacesize;               /* size of fromspace and tospace */
/* private declarations for copying collection 166c */
static Value *hp, *heaplimit;                /* used for every allocation */
/* private declarations for copying collection 167b */
static void scanvaluelist(Valuelist *vl);
static void scanenv      (Env *env);
static void scanexp      (Exp *exp);
static void scanexplist  (Explist *el);
static void scantoplevel (Toplevel *top);
static void scanvalue    (Value *vp);
static void scanroot     (Root r);
/* private declarations for copying collection 169a */
#define isinspace(LOC, SPACE) ((SPACE) <= (LOC) && (LOC) < (SPACE) +\
                                                                  semispacesize)
static Value *forward(Value *p);
/* private declarations for copying collection 569e */
static void collect(void);
/* copy.c 167a */
int nalloc;   /* OMIT */
Value*
allocloc(void)
{
    if (hp == heaplimit)
        collect();
    assert(hp < heaplimit);
    nalloc++;   /* OMIT */
    *hp = mkInvalid("allocated but uninitialized");   /* OMIT */
    return hp++;
}
/* copy.c 167e */
static void
scanenv(Env *env)
{
    for (; env; env = env->tl)
        env->loc = forward(env->loc);
}
/* copy.c 168a */
static void
scanvalue(Value *vp)
{
    switch (vp->ty) {
    case NIL:
    case BOOL:
    case NUM:
    case SYM:
        return;
    case PAIR:
        vp->u.pair.car = forward(vp->u.pair.car);
        vp->u.pair.cdr = forward(vp->u.pair.cdr);
        return;
    case CLOSURE:
        scanexp(vp->u.closure.lambda.body);
        scanenv(vp->u.closure.env);
        return;
    case PRIMITIVE:
        return;
    default:
        assert(0);
        return;
    }
}
/* copy.c 168b */
static Value*
forward(Value *p)
{
    if (isinspace(p, tospace)) { /* already in to space; must belong to scanned
                                                                         root */
        return p;
    } else {
        assert(isinspace(p, fromspace));
        /* forward pointer [[p]] and return the result 162b */
        if (p->ty == FORWARD) {            /* forwarding pointer */
            assert(isinspace(p->u.forward, tospace));   /* OMIT */
            return p->u.forward;
        } else {
            assert(isinspace(hp, tospace)); /* there is room */   /* OMIT */
            *hp = *p;
            *p  = mkForward(hp);        /* overwrite *p with a new forwarding
                                                                      pointer */
            return hp++;
        }
    }
}
/* copy.c 553 */
static void
scanexp(Exp *e)
{
    switch (e->ty) {
    case LITERAL:
        scanvalue(&e->u.literal);
        break;
    case VAR:
        break;
    case IFX:
        scanexp(e->u.ifx.cond);
        scanexp(e->u.ifx.true);
        scanexp(e->u.ifx.false);
        break;
    case WHILEX:
        scanexp(e->u.whilex.cond);
        scanexp(e->u.whilex.body);
        break;
    case BEGIN:
        scanexplist(e->u.begin);
        break;
    case SET:
        scanexp(e->u.set.exp);
        break;
    case LETX:
        scanexplist(e->u.letx.el);
        scanexp(e->u.letx.body);
        break;
    case LAMBDAX:
        scanexp(e->u.lambdax.body);
        break;
    case APPLY:
        scanexp(e->u.apply.fn);
        scanexplist(e->u.apply.actuals);
        break;
    default:
        assert(0);
    }
}
/* copy.c 554a */
static void
scantoplevel(Toplevel *t)
{
    if (t != NULL)
        switch (t->ty) {
        case EXP:
            scanexp(t->u.exp);
            break;
        case DEFINE:
            scanexp(t->u.define.lambda.body);
            break;
        case VAL:
            scanexp(t->u.val.exp);
            break;
        case USE:
            break;
        default:
            assert(0);
        }
}
/* copy.c 554b */
static void
scanroot(Root r)
{
    switch (r.ty) {
    case STACKVALUEROOT:
        scanvalue(r.u.stackvalueroot);
        return;
    case HEAPLOCROOT:
        *r.u.heaplocroot = forward(*r.u.heaplocroot);
        return;
    case ENVROOT:
        scanenv(*r.u.envroot);
        return;
    case EXPROOT:
        scanexp(*r.u.exproot);
        return;
    case VALUELISTROOT:
        scanvaluelist(*r.u.valuelistroot);
        return;
    case TOPLEVELROOT:
        scantoplevel(*r.u.toplevelroot);
        return;
    default:
        assert(0);
    }
}
/* copy.c 555a */
static void
scanvaluelist(Valuelist *vl)
{
    for (; vl; vl = vl->tl)
        scanvalue(&vl->hd);
}
/* copy.c 555b */
static void
scanexplist(Explist *el)
{
    for (; el; el = el->tl)
        scanexp(el->hd);
}
/* copy.c */
/* you need to redefine these functions */
void collect(void) { assert(0); }
void printfinalstats(void) { assert(0); }
