/* eval.c 557a */
#include "all.h"

/* eval.c declarations 559d */
static Valuelist *evallist(Explist *el, Env *env);
/* eval.c 557b */
Value
eval(Exp *e, Env *env)
{
    switch (e->ty) {
    case LITERAL:
        /* evaluate [[e->u.literal]] and return 557c */
        return e->u.literal;
    case VAR:   
        /* evaluate [[e->u.var]] and return 557d */
        if (find(e->u.var, env) == NULL)
            error("variable %n not found", e->u.var);
        return *find(e->u.var, env);
    case SET:
        /* evaluate [[e->u.set]] and return 558a */
        {
            Value v;

            if (find(e->u.set.name, env) == NULL)
                error("set unbound variable %n", e->u.set.name);
            v = eval(e->u.set.exp, env);
            return *find(e->u.set.name, env) = v;
        }
    case IFX:
        /* evaluate [[e->u.if]] and return 562a */
        if (istrue(eval(e->u.ifx.cond, env)))
            return eval(e->u.ifx.true, env);
        else
            return eval(e->u.ifx.false, env);
    case WHILEX:
        /* evaluate [[e->u.while]] and return 562b */
        while (istrue(eval(e->u.whilex.cond, env)))
            eval(e->u.whilex.body, env);
        return false;
    case BEGIN:
        /* evaluate [[e->u.begin]] and return 562c */
        {
            Explist *el;
            Value v;

            v = false;
            for (el = e->u.begin; el; el = el->tl)
                v = eval(el->hd, env);
            return v;
        }
    case LETX:
        /* evaluate let expression and return 560b */
        {
            Value rv;

            pushroot(mkEnvroot(&env));
            switch (e->u.letx.let) {
            case LET:
                if (duplicatename(e->u.letx.nl) != NULL)
                    error("name %n appears twice in let", duplicatename(e->
                                                                    u.letx.nl));
                /* extend [[env]] by simultaneously binding [[el]] to [[nl]] 561
                                                                            a */
                {
                    Valuelist *vl;

                    vl = evallist(e->u.letx.el, env);
                    pushroot(mkValuelistroot(&vl));
                    env = bindalloclist(e->u.letx.nl, vl, env);
                    poproot(mkValuelistroot(&vl));
                }
                break;
            case LETSTAR:
                /* extend [[env]] by sequentially binding [[el]] to [[nl]] 561b
                                                                              */
                {
                    Namelist *nl;
                    Explist *el;

                    for (nl = e->u.letx.nl, el = e->u.letx.el; nl && el; nl = nl
                                                              ->tl, el = el->tl)
                        env = bindalloc(nl->hd, eval(el->hd, env), env);
                    assert(nl == NULL && el == NULL);
                }
                break;
            case LETREC:
                if (duplicatename(e->u.letx.nl) != NULL)
                    error("name %n appears twice in letrec", duplicatename(e->
                                                                    u.letx.nl));
                /* extend [[env]] by recursively binding [[el]] to [[nl]] 561c
                                                                              */
                {
                    Namelist *nl;
                    Explist *el;
                    Value v;

                    for (nl = e->u.letx.nl; nl; nl = nl->tl)
                        env = bindalloc(nl->hd, mkNil(), env);

                    for (nl = e->u.letx.nl, el = e->u.letx.el; nl && el; nl = nl
                                                              ->tl, el = el->tl)
                    {
                        v = eval(el->hd, env);
                        *find(nl->hd, env) = v;
                    }
                }
                break;
            default:
                assert(0);
            }

            rv = eval(e->u.letx.body, env);
            poproot(mkEnvroot(&env));
            return rv;
        }
    case LAMBDAX:
        /* evaluate [[e->u.lambdax]] and return 558b */
        if (duplicatename(e->u.lambdax.formals) != NULL)
            error("formal parameter %n appears twice in lambda",
                  duplicatename(e->u.lambdax.formals));
        return mkClosure(e->u.lambdax, env);
    case APPLY:
        /* evaluate [[e->u.apply]] and return 559a */
        {
            Value f, rv;
            Valuelist *vl;

            f = eval(e->u.apply.fn, env);
            pushroot(mkStackvalueroot(&f));
            vl = evallist(e->u.apply.actuals, env);
            pushroot(mkValuelistroot(&vl));

            switch (f.ty) {
            case PRIMITIVE:
                /* apply [[f.u.primitive]], storing the result in [[rv]] 559b */
                rv = f.u.primitive.function(e, f.u.primitive.tag, vl);
                break;
            case CLOSURE:
                /* apply [[f.u.closure]], storing the result in [[rv]] 559c */
                {
                    Env *env;
                    Namelist *nl;

                    nl = f.u.closure.lambda.formals;
                    checkargc(e, lengthNL(nl), lengthVL(vl));

                    env = bindalloclist(nl, vl, f.u.closure.env);
                    pushroot(mkEnvroot(&env));
                    rv = eval(f.u.closure.lambda.body, env);
                    poproot(mkEnvroot(&env));
                }
                break;
            default:
                error("%e evaluates to non-function %v in %e", e->u.apply.fn, f,
                                                                             e);
            }
            poproot(mkValuelistroot(&vl));
            poproot(mkStackvalueroot(&f));
            return rv;
        }
    }
    assert(0);
    return false;
}
/* eval.c 560a */
static Valuelist*
evallist(Explist *el, Env *env)
{
    Value v;
    Valuelist *rv;

    if (el == NULL) {
        return NULL;
    } else {
        v = eval(el->hd, env);      /* enforce uScheme's order of evaluation */
        pushroot(mkStackvalueroot(&v));
        rv = mkVL(v, evallist(el->tl, env));
        poproot(mkStackvalueroot(&v));
        return rv;
    }
}
/* eval.c 562d */
Env*
topeval(Toplevel *t, Env *env, int echo)
{
    switch (t->ty) {
    case EXP:
        /* evaluate top-level expression and return [[env]] 563a */
        if (echo)
            print("%v\n", eval(t->u.exp, env));
        else
            eval(t->u.exp, env);
        return env;
    case DEFINE:
        /* evaluate function definition and return new environment 563c */
        if (duplicatename(t->u.define.lambda.formals) != NULL)
            error(
               "formal parameter %n appears twice in definition of function %n",
                  duplicatename(t->u.define.lambda.formals), t->u.define.name);
        return topeval(mkVal(t->u.define.name, mkLambdax(t->u.define.lambda)),
                                                                     env, echo);
    case VAL:
        /* evaluate [[val]] binding and return new environment 563b */
        {
            Value v;

            if (find(t->u.val.name, env) == NULL)
                env = bindalloc(t->u.val.name, mkNil(), env);
            pushroot(mkEnvroot(&env));
            v = eval(t->u.val.exp, env);
            poproot(mkEnvroot(&env));
            *find(t->u.val.name, env) = v;
            if (echo) {
                if (t->u.val.exp->ty == LAMBDAX)
                    print("%n\n", t->u.val.name);
                else
                    print("%v\n", v);
            }
            return env;
        }
    case USE:
        /* read in a file and return new environment 563d */
        {
            FILE *fin;
            const char *filename;

            filename = nametostr(t->u.use);
            fin = fopen(filename, "r");
            if (fin == NULL)
                error("cannot open file \"%s\"", filename);
            pushroot(mkEnvroot(&env));
            readevalprint(topreader(filereader(filename, fin), 0), &env, 1);
            poproot(mkEnvroot(&env));
            fclose(fin);
            return env;
        }
    default:
        assert(0);
        return NULL;
    }
}
/* eval.c 564a */
void
readevalprint(Topreader *reader, Env **envp, int echo)
{
    Toplevel *t;

    pushroot(mkToplevelroot(&t));
    t = NULL; /* don't allow garbage on stack during first call to [[readtop]]
                                                                              */
    while ((t = readtop(reader)))
        *envp = topeval(t, *envp, echo);
    poproot(mkToplevelroot(&t));
}
