/* prim.c 113c */
#include "all.h"

static Primitive arith, binary, unary;
/* prim.c 114a */
enum {
  #define xx(NAME, TAG, FUNCTION) TAG,
  #include "prim.h"
  #undef xx
  UNUSED_TAG
};
/* prim.c 114b */
Env*
primenv(void)
{
    Env *env;

    env = NULL;
    pushroot(mkEnvroot(&env));
    #define xx(NAME, TAG, FUNCTION) \
        env = bindalloc(strtoname(NAME), mkPrimitive(TAG, FUNCTION), env);
    #include "prim.h"
    #undef xx
    poproot(mkEnvroot(&env));
    return env;
}
/* prim.c 114d */
static int
projectint(Exp *e, Value v)
{
    if (v.ty != NUM)
        error("in %e, expected an integer, but got %v", e, v);
    return v.u.num;
}
/* prim.c 115a */
static Value
bool(int i)
{
    return i ? true : false;
}
/* prim.c 115b */
static int
divide(int n, int m)
{
    if (n >= 0)
        if (m >= 0)
            return n / m;
        else
            return -(( n - m - 1) / -m);
    else
        if (m >= 0)
            return -((-n + m - 1) /  m);
        else
            return -n / -m;
}
/* prim.c 116a */
static Value
arith(Exp *e, int tag, Valuelist *args)
{
    int n, m;

    checkargc(e, 2, lengthVL(args));
    n = projectint(e, nthVL(args, 0));
    m = projectint(e, nthVL(args, 1));

    switch (tag) {
    case PLUS:
        return mkNum(n + m);
    case MINUS:
        return mkNum(n - m);
    case TIMES:
        return mkNum(n * m);
    case DIV:
        if (m==0)
            error("division by zero");
        return mkNum(divide(n, m));
    case LT:
        return bool(n < m);
    case GT:
        return bool(n > m);
    default:
        assert(0);
        return false;
    }
}
/* prim.c 117a */
static Value
binary(Exp *e, int tag, Valuelist *args)
{
    Value v, w;

    checkargc(e, 2, lengthVL(args));
    v = nthVL(args, 0);
    w = nthVL(args, 1);

    switch (tag) {
    case CONS:
        /* return cons cell containing [[v]] and [[w]] 117b */
        {
            Value *l;
            Value rv;

            pushroot(mkStackvalueroot(&v));
            pushroot(mkStackvalueroot(&w));
            l = allocate(v);
            pushroot(mkHeaplocroot(&l));
            rv = mkPair(l, allocate(w));
            poproot(mkHeaplocroot(&l));
            poproot(mkStackvalueroot(&w));
            poproot(mkStackvalueroot(&v));
            return rv;
        }
    case EQ: 
        /* return [[(= v w)]] 117c */
        if (v.ty != w.ty)
            return false;

        switch (v.ty) {
        case NUM:
            return bool(v.u.num  == w.u.num);
        case BOOL:
            return bool(v.u.bool == w.u.bool);
        case SYM:
            return bool(v.u.sym  == w.u.sym);
        case NIL:
            return true;
        default:
            return false;
        }
    }
    assert(0);
    return false; /* not reached */
}
/* prim.c 119 */
static Value
unary(Exp *e, int tag, Valuelist *args)
{
    Value v;

    checkargc(e, 1, lengthVL(args));
    v = nthVL(args, 0);
    switch (tag) {
    case NULLP:
        return bool(v.ty == NIL);
    case BOOLEANP:
        return bool(v.ty == BOOL);
    case NUMBERP:
        return bool(v.ty == NUM);
    case SYMBOLP:
        return bool(v.ty == SYM);
    case PAIRP:
        return bool(v.ty == PAIR);
    case PROCEDUREP:
        return bool(v.ty == CLOSURE || v.ty == PRIMITIVE);
    case CAR:
        if (v.ty != PAIR)
            error("car applied to non-pair %v in %e", v, e);
        return *v.u.pair.car;
    case CDR:
        if (v.ty != PAIR)
            error("cdr applied to non-pair %v in %e", v, e);
        return *v.u.pair.cdr;
    case PRINT:
        print("%v\n", v);
        return v;
    case ERROR:
        error("%v", v);
        return v;
    default:
        assert(0);
        return false; /* not reached */
    }
}
