#include "all.h"
Toplevel*
mkExp(Exp *x)
{
    Toplevel *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = EXP;
    n->u.exp = x;
    return n;
}

Toplevel*
mkDefine(Name *name, Lambda lambda)
{
    Toplevel *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = DEFINE;
    n->u.define.name = name;
    n->u.define.lambda = lambda;
    return n;
}

Toplevel*
mkVal(Name *name, Exp *exp)
{
    Toplevel *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = VAL;
    n->u.val.name = name;
    n->u.val.exp = exp;
    return n;
}

Toplevel*
mkUse(Name *x)
{
    Toplevel *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = USE;
    n->u.use = x;
    return n;
}

Exp*
mkLiteral(Value x)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = LITERAL;
    n->u.literal = x;
    return n;
}

Exp*
mkVar(Name *x)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = VAR;
    n->u.var = x;
    return n;
}

Exp*
mkSet(Name *name, Exp *exp)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = SET;
    n->u.set.name = name;
    n->u.set.exp = exp;
    return n;
}

Exp*
mkIfx(Exp *cond, Exp *true, Exp *false)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = IFX;
    n->u.ifx.cond = cond;
    n->u.ifx.true = true;
    n->u.ifx.false = false;
    return n;
}

Exp*
mkWhilex(Exp *cond, Exp *body)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = WHILEX;
    n->u.whilex.cond = cond;
    n->u.whilex.body = body;
    return n;
}

Exp*
mkBegin(Explist *x)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = BEGIN;
    n->u.begin = x;
    return n;
}

Exp*
mkLetx(Letkeyword let, Namelist *nl, Explist *el, Exp *body)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = LETX;
    n->u.letx.let = let;
    n->u.letx.nl = nl;
    n->u.letx.el = el;
    n->u.letx.body = body;
    return n;
}

Exp*
mkLambdax(Lambda x)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = LAMBDAX;
    n->u.lambdax = x;
    return n;
}

Exp*
mkApply(Exp *fn, Explist *actuals)
{
    Exp *n;

    n = malloc(sizeof *n);
    assert(n != NULL);

    n->ty = APPLY;
    n->u.apply.fn = fn;
    n->u.apply.actuals = actuals;
    return n;
}

