/* ms.c 156a */
#include "all.h"

/* private declarations for mark-and-sweep collection 156b */
typedef struct Mvalue Mvalue;
struct Mvalue {
    Value v;
    unsigned live:1;
};
/* private declarations for mark-and-sweep collection 156c */
enum growth_unit { GROWTH_UNIT = 24 };
typedef struct Arena Arena;
struct Arena {
    Mvalue pool[GROWTH_UNIT];
    Arena *tl;
};
/* private declarations for mark-and-sweep collection 156d */
Arena *arenalist, *curarena;
Mvalue *hp, *heaplimit;
/* private declarations for mark-and-sweep collection 158a */
static void visitheaploc      (Value *loc);
static void visitstackvalue   (Value v);
static void visitvaluechildren(Value v);
static void visitenv          (Env *env);
static void visitvaluelist    (Valuelist *vl);
static void visitexp          (Exp *exp);
static void visitexplist      (Explist *el);
static void visittoplevel     (Toplevel *top);
static void visitroot         (Root r);
/* private declarations for mark-and-sweep collection 677a */
static int nalloc;              /* total number of allocations */
static int ncollections;        /* total number of collections */
static int nmarks;              /* total number of cells marked */
/* ms.c 157a */
static void
makecurrent(Arena *a)
{
    assert(a != NULL);
    curarena = a;
    hp = &a->pool[0];
    heaplimit = &a->pool[GROWTH_UNIT];
}
/* ms.c 157b */
static int heapsize;            /* OMIT */
static void
addarena(void)
{
    Arena *a;

    a = calloc(1, sizeof(*a));
    assert(a != NULL);

    if (arenalist == NULL) {
        arenalist = a;
    } else {
        assert(curarena != NULL && curarena->tl == NULL);
        curarena->tl = a;
    }
    makecurrent(a);
    heapsize += GROWTH_UNIT;   /* OMIT */
}
/* ms.c 158b */
Value*
allocloc(void)
{
    if (hp == heaplimit)
        addarena();
    assert(hp < heaplimit);
    return &(hp++)->v;
}
/* ms.c 159a */
static void
visitenv(Env *env)
{
    for (; env; env = env->tl)
        visitheaploc(env->loc);
}
/* ms.c 159b */
static void
visitheaploc(Value *loc)
{
    Mvalue *m;

    m = (Mvalue*)loc;
    if (!m->live) {
        m->live = 1;
        visitvaluechildren(m->v);
    }
}
/* ms.c */
static void 
visitstackvalue(Value v)
{
    visitvaluechildren(v);
}
/* ms.c */
static void
visitvaluechildren(Value v)
{
    switch (v.ty) {
    case NIL:
    case BOOL:
    case NUM:
    case SYM:
    case PRIMITIVE:
        return;
    case PAIR:
        visitheaploc(v.u.pair.car);
        visitheaploc(v.u.pair.cdr);
        return;
    case CLOSURE:
        visitexp(v.u.closure.lambda.body);
        visitenv(v.u.closure.env);
        return;
    default:
        assert(0);
        return;
    }
}
/* ms.c 550 */
static void
visitexp(Exp *e)
{
    switch (e->ty) {
    case LITERAL:
        visitvaluechildren(e->u.literal);
        return;
    case VAR:
        return;
    case IFX:
        visitexp(e->u.ifx.cond);
        visitexp(e->u.ifx.true);
        visitexp(e->u.ifx.false);
        return;
    case WHILEX:
        visitexp(e->u.whilex.cond);
        visitexp(e->u.whilex.body);
        return;
    case BEGIN:
        visitexplist(e->u.begin);
        return;
    case SET:
        visitexp(e->u.set.exp);
        return;
    case LETX:
        visitexplist(e->u.letx.el);
        visitexp(e->u.letx.body);
        return;
    case LAMBDAX:
        visitexp(e->u.lambdax.body);
        return;
    case APPLY:
        visitexp(e->u.apply.fn);
        visitexplist(e->u.apply.actuals);
        return;
    default:
        assert(0);
    }
}
/* ms.c 551a */
static void
visittoplevel(Toplevel *t)
{
    if (t == NULL)
        return;
    else
        switch (t->ty) {
        case EXP:
            visitexp(t->u.exp);
            return;
        case VAL:
            visitexp(t->u.val.exp);
            return;
        case DEFINE:
            visitexp(t->u.define.lambda.body);
            return;
        case USE:
            return;
        default:
            assert(0);
        }
}
/* ms.c 551b */
static void
visitroot(Root r)
{
    switch (r.ty) {
    case STACKVALUEROOT:
        visitstackvalue(*r.u.stackvalueroot);
        return;
    case HEAPLOCROOT:
        visitheaploc(*r.u.heaplocroot);
        return;
    case ENVROOT:
        visitenv(*r.u.envroot);
        return;
    case EXPROOT:
        visitexp(*r.u.exproot);
        return;
    case VALUELISTROOT:
        visitvaluelist(*r.u.valuelistroot);
        return;
    case TOPLEVELROOT:
        visittoplevel(*r.u.toplevelroot);
        return;
    default:
        assert(0);
    }
}
/* ms.c 552a */
static void
visitvaluelist(Valuelist *vl)
{
    for (; vl; vl = vl->tl)
        visitvaluechildren(vl->hd);
}
/* ms.c 552b */
static void
visitexplist(Explist *el)
{
    for (; el; el = el->tl)
        visitexp(el->hd);
}
/* ms.c */
/* you need to redefine these functions */
void printfinalstats(void) { assert(0); }
