/* value.c 531a */
#include "all.h"

int
istrue(Value v)
{
    return v.ty != BOOL || v.u.bool;
}

Value true, false;

void
initvalue(void)
{
    true  = mkBool(1);
    false = mkBool(0);
}
/* value.c 531b */
static int
nameinlist(Name *n, Namelist *nl)
{
    for (; nl; nl=nl->tl)
        if (n == nl->hd)
            return 1;
    return 0;
}
/* value.c 531c */
static Namelist*
addname(Name *n, Namelist *nl)
{
    if (nameinlist(n, nl))
        return nl;
    return mkNL(n, nl);
}
/* value.c 532 */
static Namelist*
addfree(Name *n, Namelist *bound, Namelist *free)
{
    if (nameinlist(n, bound))
        return free;
    return addname(n, free);
}
/* value.c 533a */
static Namelist*
freevars(Exp *e, Namelist *bound, Namelist *free)
{
    Namelist *nl;
    Explist *el;

    switch (e->ty) {
    case LITERAL:
        break;
    case VAR:
        free = addfree(e->u.var, bound, free);
        break;
    case IFX:
        free = freevars(e->u.ifx.cond, bound, free);
        free = freevars(e->u.ifx.true, bound, free);
        free = freevars(e->u.ifx.false, bound, free);
        break;
    case WHILEX:
        free = freevars(e->u.whilex.cond, bound, free);
        free = freevars(e->u.whilex.body, bound, free);
        break;
    case BEGIN:
        for (el = e->u.begin; el; el = el->tl)
            free = freevars(el->hd, bound, free);
        break;
    case SET:
        free = addfree(e->u.set.name, bound, free);
        free = freevars(e->u.set.exp, bound, free);
        break;
    case APPLY:
        free = freevars(e->u.apply.fn, bound, free);
        for (el = e->u.apply.actuals; el; el = el->tl)
            free = freevars(el->hd, bound, free);
        break;
    case LAMBDAX:
        /* let [[free]] be the free variables for [[e->u.lambdax]] 533b */
        for (nl = e->u.lambdax.formals; nl; nl = nl->tl)
            bound = addname(nl->hd, bound);
        free = freevars(e->u.lambdax.body, bound, free);
        break;
    case LETX:
        /* let [[free]] be the free variables for [[e->u.letx]] 534 */
        switch (e->u.letx.let) {
        case LET:
            for (el = e->u.letx.el; el; el = el->tl)
                free = freevars(el->hd, bound, free);
            for (nl = e->u.letx.nl; nl; nl = nl->tl)
                bound = addname(nl->hd, bound);
            free = freevars(e->u.letx.body, bound, free);
            break;
        case LETSTAR:
            for ( nl = e->u.letx.nl, el = e->u.letx.el
               ; nl && el
               ; nl = nl->tl, el = el->tl
               ) 
            {
                free  = freevars(el->hd, bound, free);
                bound = addname(nl->hd, bound);
            }
            free = freevars(e->u.letx.body, bound, free);
            break;
        case LETREC:
            for (nl = e->u.letx.nl; nl; nl = nl->tl)
                bound = addname(nl->hd, bound);
            for (el = e->u.letx.el; el; el = el->tl)
                free = freevars(el->hd, bound, free);
            free = freevars(e->u.letx.body, bound, free);
            break;
        }
        break;
    }
    return free;
}
/* value.c 535a */
static void printnonglobals(Namelist *nl, Env *env, int depth);

static void
printclosureat(Lambda lambda, Env *env, int depth) 
{
    Namelist *vars; /* free variables of lambda */

    if (depth > 0) {
        print("<%\\, {", lambda);
        vars = freevars(lambda.body, lambda.formals, NULL);
        printnonglobals(vars, env, depth - 1);
        print("}>");
    } else {
        print("<procedure>");
    }
}
/* value.c 535b */
void
printclosure(va_list *va)
{
    Lambda l;
    Env *e;

    l = va_arg(*va, Lambda);
    e = va_arg(*va, Env*);
    printclosureat(l, e, 1);
}
/* value.c 536a */
static void printvalueat(Value v, int depth);
/* helper functions for [[printvalue]] 537a */
static void
printtail(Value v, int depth)
{
    switch (v.ty) {
    case NIL:
        print(")");
        break;
    case PAIR:
        print(" ");
        printvalueat(*v.u.pair.car, depth);
        printtail(*v.u.pair.cdr, depth);
        break;
    default:
        print(" . ");
        printvalueat(v, depth);
        print(")");
        break;
    }
}
static void
printvalueat(Value v, int depth)
{
    switch (v.ty){
    case NIL:
        print("()");
        return;
    case BOOL:
        print(v.u.bool ? "#t" : "#f");
        return;
    case NUM:
        print("%d", v.u.num);
        return;
    case SYM:
        print("%n", v.u.sym);
        return;
    case PRIMITIVE:
        print("<procedure>");
        return;
    case PAIR:
        print("(");
        printvalueat(*v.u.pair.car, depth);
        printtail(*v.u.pair.cdr, depth);
        return;
    case CLOSURE:
        printclosureat(v.u.closure.lambda, v.u.closure.env, depth);
        return;
    default:
        print("<unknown v.ty=%d>", v.ty);
return;        assert(0);
    }
}
/* value.c 536b */
void
printvalue(va_list *va)
{
    printvalueat(va_arg(*va, Value), 0);
}
/* value.c 537b */
Env **globalenv;
static void
printnonglobals(Namelist *nl, Env *env, int depth)
{
    char *prefix;
    Value *loc;

    prefix = "";
    for (; nl; nl = nl->tl) {
        loc = find(nl->hd, env);
        if (loc && (globalenv == NULL || find(nl->hd, *globalenv) != loc)) {
            print("%s%n -> ", prefix, nl->hd);
            prefix = ", ";
            printvalueat(*loc, depth);
        }
    }
}
