/* {\Tt all.h} for \uscheme 102b */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <ctype.h>

/* shared type definitions 25e */
typedef struct Name Name;
typedef struct Namelist Namelist;   /* list of Name* */
/* shared type definitions 28d */
typedef struct Reader Reader;
/* shared type definitions 29a */
typedef struct Topreader Topreader;
/* shared type definitions 31e */
typedef void Printer(va_list*);
/* shared type definitions 499d */
typedef struct Parlist Parlist; /* list of Par* */
/* shared type definitions */
typedef struct Par Par;
/* type definitions for \uscheme */
typedef struct Lambda Lambda;
typedef struct Value Value;
/* type definitions for \uscheme */
typedef struct Toplevel Toplevel;
typedef struct Exp Exp;
/* type definitions for \uscheme 91b */
typedef enum Letkeyword { LET, LETSTAR, LETREC } Letkeyword;
typedef struct Explist Explist; /* list of Exp* */
/* type definitions for \uscheme 91d */
typedef struct Valuelist Valuelist;     /* list of Value */
typedef Value Primitive(Exp *e, int tag, Valuelist *vl);
/* type definitions for \uscheme 102c */
typedef struct Env Env;
/* shared structure definitions 503a */
struct Reader {
    char *buf;               /* holds the last line read */
    int nbuf;                /* size of buf */
    int line;                /* current line number */
    const char *readername;  /* identifies this reader */

    FILE *fin;               /* filereader */
    const char *s;           /* stringreader */
};
/* shared structure definitions */
typedef enum { ATOM, LIST } Party;

struct Par {
    Party ty;
    union {
        Name *atom;
        Parlist *list;
    } u;
};

/* structure definitions for \uscheme */
struct Lambda {
    Namelist *formals;
    Exp *body;
};
typedef enum { NIL, BOOL, NUM, SYM, PAIR, CLOSURE, PRIMITIVE } Valuety;

struct Value {
    Valuety ty;
    union {
        int bool;
        int num;
        Name *sym;
        struct {  Value *car; Value *cdr; } pair;
        struct {  Lambda lambda; Env *env; } closure;
        struct {  int tag; Primitive *function; } primitive;
    } u;
};

/* structure definitions for \uscheme */
typedef enum { EXP, DEFINE, VAL, USE } Toplevelty;

struct Toplevel {
    Toplevelty ty;
    union {
        Exp *exp;
        struct {  Name *name; Lambda lambda; } define;
        struct {  Name *name; Exp *exp; } val;
        Name *use;
    } u;
};

typedef enum { LITERAL, VAR, SET, IFX, WHILEX, BEGIN, LETX, LAMBDAX, APPLY }
                                                                          Expty;

struct Exp {
    Expty ty;
    union {
        Value literal;
        Name *var;
        struct {  Name *name; Exp *exp; } set;
        struct {  Exp *cond; Exp *true; Exp *false; } ifx;
        struct {  Exp *cond; Exp *body; } whilex;
        Explist *begin;
        struct {  Letkeyword let; Namelist *nl; Explist *el; Exp *body; } letx;
        Lambda lambdax;
        struct {  Exp *fn; Explist *actuals; } apply;
    } u;
};

/* structure definitions for \uscheme */
struct Explist {
    Exp *hd;
    Explist *tl;
};

struct Parlist {
    Par *hd;
    Parlist *tl;
};

struct Valuelist {
    Value hd;
    Valuelist *tl;
};

struct Namelist {
    Name *hd;
    Namelist *tl;
};

/* shared function prototypes 25f */
Name *strtoname(const char *s);
const char *nametostr(Name *n);
/* shared function prototypes 26a */
Printer printname;
/* shared function prototypes 28e */
char *readline(Reader *r, char *prompt);
/* shared function prototypes 28f */
Reader *stringreader(const char *stringname, const char *s);
Reader *filereader  (const char *filename,   FILE *fin);
/* shared function prototypes 29b */
Toplevel *readtop(Topreader *r);
/* shared function prototypes 29c */
Topreader *topreader(Reader *r, int doprompt);
/* shared function prototypes 31d */
void print (const char *fmt, ...);
void vprint(const char *fmt, va_list va);
/* shared function prototypes 31f */
void installprinter(unsigned char c, Printer *fmt);
/* shared function prototypes 32a */
Printer printpercent, printstring, printdecimal;
/* shared function prototypes 32b */
void error(const char *fmt, ...);
extern jmp_buf errorjmp;        /* longjmp here on error */
/* shared function prototypes 32c */
void checkargc(Exp *e, int expected, int actual);
/* shared function prototypes 32d */
Name *duplicatename(Namelist *names);
/* shared function prototypes 499e */
/* shared function prototypes 500a */
Printer printpar;
/* shared function prototypes */
Parlist *readparlist(Reader *r, int doquote, int doprompt);
/* shared function prototypes */
Par *mkAtom(Name*);
Par *mkList(Parlist*);
/* function prototypes for \uscheme */
Lambda mkLambda(Namelist *formals, Exp *body);
Value mkNil(void);
Value mkBool(int);
Value mkNum(int);
Value mkSym(Name*);
Value mkPair(Value *car, Value *cdr);
Value mkClosure(Lambda lambda, Env *env);
Value mkPrimitive(int tag, Primitive *function);
/* function prototypes for \uscheme */
Toplevel *mkExp(Exp*);
Toplevel *mkDefine(Name *name, Lambda lambda);
Toplevel *mkVal(Name *name, Exp *exp);
Toplevel *mkUse(Name*);
Exp *mkLiteral(Value);
Exp *mkVar(Name*);
Exp *mkSet(Name *name, Exp *exp);
Exp *mkIfx(Exp *cond, Exp *true, Exp *false);
Exp *mkWhilex(Exp *cond, Exp *body);
Exp *mkBegin(Explist*);
Exp *mkLetx(Letkeyword let, Namelist *nl, Explist *el, Exp *body);
Exp *mkLambdax(Lambda);
Exp *mkApply(Exp *fn, Explist *actuals);
/* function prototypes for \uscheme */
int      lengthEL (Explist *l);
Exp*     nthEL    (Explist *l, unsigned n);
Explist *mkEL     (Exp *hd, Explist *tl);
Printer  printexplist;
/* function prototypes for \uscheme */
int      lengthPL (Parlist *l);
Par*     nthPL    (Parlist *l, unsigned n);
Parlist *mkPL     (Par *hd, Parlist *tl);
Printer  printparlist;
/* function prototypes for \uscheme */
int        lengthVL (Valuelist *l);
Value      nthVL    (Valuelist *l, unsigned n);
Valuelist *mkVL     (Value hd, Valuelist *tl);
Printer    printvaluelist;
/* function prototypes for \uscheme */
int       lengthNL (Namelist *l);
Name*     nthNL    (Namelist *l, unsigned n);
Namelist *mkNL     (Name *hd, Namelist *tl);
Printer   printnamelist;
/* function prototypes for \uscheme 102d */
Value *find(Name *name, Env *env);
/* function prototypes for \uscheme 103a */
Env *bindalloc    (Name *name,   Value v,       Env *env);
Env *bindalloclist(Namelist *nl, Valuelist *vl, Env *env);
/* function prototypes for \uscheme 103b */
Value *allocate(Value v);
/* function prototypes for \uscheme 103c */
void initallocate(Env **envp);
/* function prototypes for \uscheme 103d */
void resetallocate(Env **envp);
/* function prototypes for \uscheme 103e */
Value true, false;
/* function prototypes for \uscheme 104a */
void initvalue(void);
/* function prototypes for \uscheme 104b */
int istrue(Value v);
/* function prototypes for \uscheme 104c */
Value eval(Exp *e, Env *rho);
Env *topeval(Toplevel *t, Env *rho, int echo);
/* function prototypes for \uscheme 104d */
void readevalprint(Topreader *r, Env **envp, int echo);
/* function prototypes for \uscheme 104e */
Env *primenv(void);
/* function prototypes for \uscheme 105a */
void printenv     (va_list*);
void printvalue   (va_list*);
void printclosure (va_list*);
void printexp     (va_list*);
void printtoplevel(va_list*);
void printlambda  (va_list*);
