/* eval.c 105b */
#include "all.h"

/* eval.c declarations 108b */
static Valuelist *evallist(Explist *el, Env *env);
/* eval.c 106a */
Value
eval(Exp *e, Env *env)
{
    switch (e->ty) {
    case LITERAL:
        /* evaluate [[e->u.literal]] and return 106b */
        return e->u.literal;
    case VAR:   
        /* evaluate [[e->u.var]] and return 106c */
        if (find(e->u.var, env) == NULL)
            error("variable %n not found", e->u.var);
        return *find(e->u.var, env);
    case SET:
        /* evaluate [[e->u.set]] and return 107a */
        if (find(e->u.set.name, env) == NULL)
            error("set unbound variable %n", e->u.set.name);
        return *find(e->u.set.name, env) = eval(e->u.set.exp, env);
    case IFX:
        /* evaluate [[e->u.ifx]] and return 111a */
        if (istrue(eval(e->u.ifx.cond, env)))
            return eval(e->u.ifx.true, env);
        else
            return eval(e->u.ifx.false, env);
    case WHILEX:
        /* evaluate [[e->u.whilex]] and return 111b */
        while (istrue(eval(e->u.whilex.cond, env)))
            eval(e->u.whilex.body, env);
        return false;
    case BEGIN:
        /* evaluate [[e->u.begin]] and return 111c */
        {
            Explist *el;
            Value v;

            v = false;
            for (el = e->u.begin; el; el = el->tl)
                v = eval(el->hd, env);
            return v;
        }
    case LETX:
        /* evaluate [[e->u.letx]] and return 109a */
        switch (e->u.letx.let) {
        case LET:
            /* if [[e->u.letx.nl]] contains a duplicate, complain of error in
                                                                 [[let]] 537d */
            if (duplicatename(e->u.letx.nl) != NULL)
                error("name %n appears twice in let", duplicatename(e->u.letx.nl
                                                                             ));
            /* extend [[env]] by simultaneously binding [[el]] to [[nl]] 109b */
            env = bindalloclist(e->u.letx.nl, evallist(e->u.letx.el, env), env);
            break;
        case LETSTAR:
            /* extend [[env]] by sequentially binding [[el]] to [[nl]] 110a */
            {
                Namelist *nl;
                Explist *el;

                for (nl = e->u.letx.nl, el = e->u.letx.el;
                     nl && el;
                     nl = nl->tl, el = el->tl)
                    env = bindalloc(nl->hd, eval(el->hd, env), env);
                assert(nl == NULL && el == NULL);
            }
            break;
        case LETREC:
            /* if [[e->u.letx.nl]] contains a duplicate, complain of error in
                                                              [[letrec]] 538a */
            if (duplicatename(e->u.letx.nl) != NULL)
                error("name %n appears twice in letrec", duplicatename(e->
                                                                    u.letx.nl));
            /* extend [[env]] by recursively binding [[el]] to [[nl]] 110b */
            {
                Namelist *nl;
                Explist *el;

                for (nl = e->u.letx.nl; nl; nl = nl->tl)    
                    env = bindalloc(nl->hd, mkNil(), env);

                for (nl = e->u.letx.nl, el = e->u.letx.el;
                     nl && el;
                     nl = nl->tl, el = el->tl)
                    *find(nl->hd, env) = eval(el->hd, env);
            }
            break;
        default:
            assert(0);
        }
        return eval(e->u.letx.body, env);
    case LAMBDAX:
        /* evaluate [[e->u.lambdax]] and return 107b */
        /* if [[e->u.lambdax.formals]] contains a duplicate, call [[error]] 537c
                                                                              */
        if (duplicatename(e->u.lambdax.formals) != NULL)
            error("formal parameter %n appears twice in lambda",
                  duplicatename(e->u.lambdax.formals));
        return mkClosure(e->u.lambdax, env);
    case APPLY:
        /* evaluate [[e->u.apply]] and return 107c */
        {
            Value f;
            Valuelist *vl;

            f  = eval(e->u.apply.fn, env);
            vl = evallist(e->u.apply.actuals, env);

            switch (f.ty) {
            case PRIMITIVE:
                /* apply [[f.u.primitive]] to [[vl]] and return the result 107d
                                                                              */
                return f.u.primitive.function(e, f.u.primitive.tag, vl);
            case CLOSURE:
                /* apply [[f.u.closure]] to [[vl]] and return the result 108a */
                {
                    Namelist *nl;

                    nl = f.u.closure.lambda.formals;
                    checkargc(e, lengthNL(nl), lengthVL(vl));

                    return eval(f.u.closure.lambda.body,
                                bindalloclist(nl, vl, f.u.closure.env));
                }
            default:
                error("%e evaluates to non-function %v in %e", e->u.apply.fn, f,
                                                                             e);
            }
        }
    }
    assert(0);
    return false;
}
/* eval.c 108c */
static Valuelist*
evallist(Explist *el, Env *env)
{
    Value v;

    if (el == NULL) {
        return NULL;
    } else {
        v = eval(el->hd, env);      /* enforce uScheme's order of evaluation */
        return mkVL(v, evallist(el->tl, env));
    }
}
