/* lex.c 506e */
#include "all.h"
#include <ctype.h>
/* lex.c 507a */
static Name*
strntoname(char *s, int n)
{
    char *t;

    t = malloc(n + 1);
    assert(t != NULL);
    strncpy(t, s, n);
    t[n] = '\0';
    return strtoname(t);
}
/* lex.c 507b */
static int
isdelim(char c, int quote)
{
    if (quote && c == '\'')
        return 1;

    return c == '\0' || c == '(' || c == ')' || c == ';' || isspace(c);
}
/* lex.c 507c */
static Name*
readname(char **ps, int quote)
{
    char *p, *q;

    p = *ps;
    for (q = p; !isdelim(*q, quote); q++)
        ;
    *ps = q;
    return strntoname(p, q - p);
}
/* lex.c 507d */
static void
skipspace(char **ps)
{
    while (isspace((unsigned char)**ps))
        (*ps)++;
}
/* lex.c 508a */
enum {
    More       = 1 << 0,
    Quote      = 1 << 1,
    Rightparen = 1 << 2
};
static Par*
readpar(char **ps, Reader *r, int flag, char *moreprompt)
{
    if (*ps == NULL)
        return NULL;

    skipspace(ps);

    switch (**ps) {
    case '\0':
    case ';':
        /* readpar end of line and return 508b */
        if ((flag & More) && (*ps = readline(r, moreprompt)) != NULL)
            return readpar(ps, r, flag, moreprompt);
        *ps = NULL;
        return NULL;
    case ')':
        /* readpar right parenthesis and return 508c */
        if ((flag & Rightparen) == 0)
            error("unexpected right parenthesis in %s, line %d", r->readername,
                                                                       r->line);
        return NULL;
    case '(':
        /* readpar left parenthesis and return 509a */
        {
            Par *p, *q;
            Parlist **ppl;

            p   = mkList(NULL);
            ppl = &p->u.list;

            (*ps)++;
            while ((q = readpar(ps, r, flag | More | Rightparen, moreprompt))) {
                *ppl = mkPL(q, NULL);
                ppl  = &(*ppl)->tl;
            }
            if (*ps == NULL)
                error("premature end of file reading list (missing right paren)"
                                                                              );

            assert(**ps == ')');
            (*ps)++;    /* past right parenthesis */

            return p;
        }
    default:
        if ((flag & Quote) && **ps == '\'') {
            /* readpar quoted expression and return 509b */
            {
                Par *p;

                (*ps)++;
                p = readpar(ps, r, More|Quote, moreprompt);
                return mkList(mkPL(mkAtom(strtoname("quote")), mkPL(p, NULL)));
            }
        } else {
            /* readpar name and return 509c */
            return mkAtom(readname(ps, flag & Quote));
        }
    }   
}
/* lex.c 510a */
Parlist*
readparlist(Reader *r, int doquote, int doprompt)
{
    char *s;
    Par *p;
    Parlist *pl, **ppl;

    pl  = NULL;
    ppl = &pl;

    while (pl == NULL) {
        if ((s = readline(r, doprompt ? "-> " : "")) == NULL)
            return NULL;

        while ((p = readpar(&s, r, doquote ? Quote : 0, doprompt ? "   " : "")))
                                                                               {
            *ppl = mkPL(p, NULL);
            ppl  = &(*ppl)->tl;
        }
    }
    return pl;
}
/* lex.c 510b */
void
printpar(va_list *va)
{
    Par *p;

    p = va_arg(*va, Par*);
    if (p == NULL) {
        print("<null>");
        return;
    }

    switch (p->ty){
    case ATOM:
        print("%n", p->u.atom);
        break;
    case LIST:
        print("(%P)", p->u.list);
        break;
    }
}
