/*   Printers                                   */
/*                                              */
/* <ast.c>=                                     */
#include "all.h"

void
printexp(va_list *va)
{
    Exp *e;

    e = va_arg(*va, Exp*);
    if (e == NULL) {
        print("<null>");
        return;
    }

    switch (e->ty){
    case LITERAL:
        print("%v", e->u.literal);
        break;
    case VAR:
        print("%n", e->u.var);
        break;
    case SET:
        print("(set %n %e)", e->u.set.name, e->u.set.exp);
        break;
    case IFX:
        print("(if %e %e %e)", e->u.ifx.cond, e->u.ifx.true, e->u.ifx.false);
        break;
    case WHILEX:
        print("(while %e %e)", e->u.whilex.cond, e->u.whilex.exp);
        break;
    case BEGIN:
        print("(begin%s%E)", e->u.begin?" ":"", e->u.begin);
        break;
    case APPLY:
        print("(%n%s%E)", e->u.apply.name,
                      e->u.apply.actuals?" ":"", e->u.apply.actuals);
        break;
    }
}
/* <ast.c>=                                     */
void
printtoplevel(va_list *va)
{
    Toplevel *t;

    t = va_arg(*va, Toplevel*);
    if (t == NULL) {
        print("<null>");
        return;
    }

    switch (t->ty) {
    case EXP:
        print("%e", t->u.exp);
        break;
    case DEFINE:
        print("(define %n (%N) %e)", t->u.define.name,
                      t->u.define.userfun.formals,
              t->u.define.userfun.body);
        break;
    case VAL:
        print("(val %n %e)", t->u.val.name, t->u.val.exp);
        break;
    case USE:
        print("(use %n)", t->u.use);
        break;
    default:
        assert(0);
    }
}
