/*   Implementing [[main]]                      */
/*                                              */
/* The [[main]] program recognizes the          */
/* -q option (``quiet,'' to suppress prompts),  */
/* initializes the interpreter, and runs the    */
/* read-eval-print loop.                        */
/* <impcore.c>=                                 */
#include "all.h"

int
main(int argc, char *argv[])
{
    Funenv *functions;
    Valenv *globals;
    Topreader *input;
    int doprompt;

    doprompt = (argc <= 1) || (strcmp(argv[1], "-q") != 0);

    /* The printers are distributed throughout the  */
    /* interpreter, but they are all installed      */
    /* here.                                        */
    /* <install conversion specifications in print module>= */
    installprinter('d', printdecimal);
    installprinter('e', printexp);
    installprinter('E', printexplist);
    installprinter('f', printfun);
    installprinter('n', printname);
    installprinter('N', printnamelist);
    installprinter('p', printpar);
    installprinter('P', printparlist);
    installprinter('s', printstring);
    installprinter('t', printtoplevel);
    installprinter('v', printvalue);
    installprinter('V', printvaluelist);
    functions = mkFunenv(NULL, NULL);
    globals   = mkValenv(NULL, NULL);
    /* The initial basis includes both primitives   */
    /* and user-defined functions. We install the   */
    /* primitives first.                            */
    /* <install the initial basis in [[functions]]>= */
    {
        static char *prims[] = { "+", "-", "*", "/", "<", ">", "=", "print", 0 }
                                                                               ;
        char **p;
        Name *n;

        for (p=prims; *p; p++) {
            n = strtoname(*p);
            bindfun(n, mkPrimitive(n), functions);
        }
    }
    /* <install the initial basis in [[functions]]>= */
    {
        /* We represent the user-defined part of the    */
        /* initial basis as a single string, which is   */
        /* interpreted by [[readevalprint]]. These      */
        /* functions also appear in Figure [->] on      */
        /* page [->], from which this code is derived   */
        /* automatically.                               */
        /* <C representation of initial basis for {\impcore}>= */
        /* <C representation of initial basis for {\impcore}>= */
        const char *basis=
            "(define and (b c) (if b c b))\n"
            "(define or  (b c) (if b b c))\n"
            "(define not (b)   (if b 0 1))\n"
            "(define <= (x y) (not (> x y)))\n"
            "(define >= (x y) (not (< x y)))\n"
            "(define != (x y) (not (= x y)))\n"
            "(define mod (m n) (- m (* n (/ m n))))\n";

        if (setjmp(errorjmp))
            assert(0); /* fail if error in basis */
        readevalprint(topreader(stringreader("intial basis", basis), 0),
                      globals, functions, 0);
    }

    input = topreader(filereader("standard input", stdin), doprompt);
    while (setjmp(errorjmp))
        ;
    readevalprint(input, globals, functions, 1);
    return 0;
}
