/*   Error interface                            */
/*                                              */
/* Function [[error]] prints a message, then    */
/* jumps to [[errorjmp]].                       */
/* <error.c>=                                   */
#include "all.h"

jmp_buf errorjmp;

void
error(const char *fmt, ...)
{
    va_list va;

    print("error: ");
    va_start(va, fmt);
    vprint(fmt, va);
    va_end(va);
    print("\n");
    longjmp(errorjmp, 1);   
}
/* [[checkargc]] checks number of arguments.    */
/*                                              */
/* <error.c>=                                   */
void
checkargc(Exp *e, int expected, int actual)
{
    if (expected != actual)
        error("expected %d but found %d argument%s in %e",
              expected, actual, actual == 1 ? "" : "s", e);
}
/* Function [[duplicatename]] returns duplicate */
/* names from a list. It is used to detect      */
/* duplicate names in lists of formal           */
/* parameters. Its cost is quadratic in the     */
/* number of parameters, which should be very   */
/* fast for any reasonable function.            */
/* <error.c>=                                   */
Name *duplicatename(Namelist *nl) 
{
    Name *n;
    Namelist *tail;
      
    if (nl != NULL) {
        n = nl->hd;
        for (tail = nl->tl; tail; tail = tail->tl)
            if (n == tail->hd)
                return n;
        return duplicatename(nl->tl);
    }
    return NULL;
}
