/* <env.c>=                                     */
#include "all.h"
/* We look up a name by following [[tl]]        */
/* pointers. [*]                                */
/* <env.c>=                                     */
Value*
find(Name *name, Env *env)
{
    for (; env; env = env->tl)
        if (env->name == name)
            return env->loc;
    return NULL;
}
/* Function [[bindalloc]] always creates a new  */
/* environment with a new binding. There is     */
/* never any mutation.                          */
/* <env.c>=                                     */
Env*
bindalloc(Name *name, Value val, Env *env)
{
    Env *newenv;

    newenv = malloc(sizeof(*newenv));
    assert(newenv != NULL);

    newenv->name = name;
    newenv->loc  = allocate(val);
    newenv->tl   = env;
    return newenv;
}
/* We include [[printenv]] in case it helps you */
/* debug your code.                             */
/* <env.c>=                                     */
void
printenv(va_list *va) 
{
    Env *env;
    char *prefix;

    env = va_arg(*va, Env*);

    print("{");
    prefix = " ";
    for (; env; env = env->tl) {
        print("%s%n -> %v", prefix, env->name, *env->loc);
        prefix = ", ";
    }
    print(" }");
}
/* This version of [[bindalloclist]] needs to   */
/* push its [[env]] because it modifies it.     */
/* Pointers returned by earlier calls to        */
/* [[bindalloc]] must be reachable from the     */
/* root stack at subsequent calls to            */
/* [[bindalloc]].                               */
/* <env.c>=                                     */
Env*
bindalloclist(Namelist *nl, Valuelist *vl, Env *env)
{
    pushroot(mkEnvroot(&env));
    for (; nl && vl; nl = nl->tl, vl = vl->tl)
        env = bindalloc(nl->hd, vl->hd, env);
    poproot(mkEnvroot(&env));
    return env;
}
/* For [[bindalloc]], [[val]] is a parameter    */
/* passed by value, so we have a fresh copy of  */
/* it. It contains [[Value*]] pointers, so you  */
/* might think it needs to be on the root stack */
/* for the copying collector (so that the       */
/* pointers can be updated if necessary). But   */
/* by the time we get to [[allocate]], our copy */
/* of [[val]] is dead--only [[allocate]]'s      */
/* private copy matters. It is therefore safe   */
/* to reuse the implementation from             */
/* Chapter [->].                                */
/* <env.c>=                                     */
