/*   Expressions                                */
/*                                              */
/* Here is the (boring) code that prints        */
/* abstract-syntax trees.                       */
/* <ast.c>=                                     */
#include "all.h"
/* <ast.c>=                                     */
void
printtoplevel(va_list *va)
{
    Toplevel *t;

    t = va_arg(*va, Toplevel*);
    if (t == NULL) {
        print("<null>");
        return;
    }

    switch (t->ty) {
    case EXP:
        print("%e", t->u.exp);
        break;
    case DEFINE:
        print("(define %n %\\)", t->u.define.name, t->u.define.lambda);
        break;
    case VAL:
        print("(val %n %e)", t->u.val.name, t->u.val.exp);
        break;
    case USE:
        print("(use %n)", t->u.use);
        break;
    default:
        assert(0);
    }
}
/* <ast.c>=                                     */
static void
printlet(Exp *let)
{
    Namelist *nl;
    Explist *el;

    switch (let->u.letx.let) {
    case LET:
        print("(let (");
        break;
    case LETSTAR:
        print("(let* (");
        break;
    case LETREC:
        print("(letrec (");
        break;
    default:
        assert(0);
    }
    for (nl = let->u.letx.nl, el = let->u.letx.el; 
         nl && el;
         nl = nl->tl, el = el->tl)
        print("(%n %e)%s", nl->hd, el->hd, nl->tl?" ":"");
    print(") %e)", let->u.letx.body);
}   
/* <ast.c>=                                     */
void
printexp(va_list *va)
{
    Exp *e;

    e = va_arg(*va, Exp*);
    if (e == NULL) {
        print("<null>");
        return;
    }

    switch (e->ty) {
    case LITERAL:
        if (e->u.literal.ty == NUM)
            print("%v", e->u.literal);
        else
            print("'%v", e->u.literal);
        break;
    case VAR:
        print("%n", e->u.var);
        break;
    case IFX:
        print("(if %e %e %e)", e->u.ifx.cond, e->u.ifx.true, e->u.ifx.false);
        break;
    case WHILEX:
        print("(while %e %e)", e->u.whilex.cond, e->u.whilex.body);
        break;
    case BEGIN:
        print("(begin%s%E)", e->u.begin ? " " : "", e->u.begin);
        break;
    case SET:
        print("(set %n %e)", e->u.set.name, e->u.set.exp);
        break;
    case LETX:
        printlet(e);
        break;
    case LAMBDAX:
        print("%\\", e->u.lambdax);
        break;
    case APPLY:
        print("(%e%s%E)", e->u.apply.fn,
              e->u.apply.actuals ? " " : "", e->u.apply.actuals);
        break;
    }
}
/* <ast.c>=                                     */
void
printlambda(va_list *va)
{
    Lambda l;

    l = va_arg(*va, Lambda);
    print("(lambda (%N) %e)", l.formals, l.body);
}
