/*   Implementation of the interpreter's [[main]] */
/*   procedure                                  */
/*                                              */
/* As in the Impcore interpreter, [[main]]      */
/* processes arguments, initializes the         */
/* interpreter, and runs the read-eval-print    */
/* loop.                                        */
/* <scheme.c>=                                  */
#include "all.h"

int
main(int argc, char *argv[])
{
    Env *env;
    Topreader *input;
    int doprompt;

    doprompt = (argc <= 1) || (strcmp(argv[1], "-q") != 0);

    initvalue();
    initallocate(&env);
    /* We have many printers.                       */
    /*                                              */
    /* <install printers>=                          */
    installprinter('c', printclosure);
    installprinter('d', printdecimal);
    installprinter('e', printexp);
    installprinter('E', printexplist);
    installprinter('\\', printlambda);
    installprinter('n', printname);
    installprinter('N', printnamelist);
    installprinter('p', printpar);
    installprinter('P', printparlist);
    installprinter('r', printenv);
    installprinter('s', printstring);
    installprinter('t', printtoplevel);
    installprinter('v', printvalue);
    installprinter('V', printvaluelist);
    installprinter('%', printpercent);
    env = primenv();
    /* As in the Impcore interpreter, the           */
    /* C representation of the initial basis is     */
    /* generated automatically from code in         */
    /* [[<<additions to the micro-Scheme initial    */
    /* basis>>]].                                   */
    /* <install into [[env]] the additions to the initial basis>= */
    {
        /* <C representation of initial basis for {\uscheme}>= */
        const char *basis=
            "(define caar (l) (car (car l)))\n"
            "(define cadr (l) (car (cdr l)))\n"
            "(define cdar (l) (cdr (car l)))\n"
            "(define list1 (x)     (cons x '()))\n"
            "(define list2 (x y)   (cons x (list1 y)))\n"
            "(define list3 (x y z) (cons x (list2 y z)))\n"
            "(define length (l)\n"
            "  (if (null? l) 0\n"
            "    (+ 1 (length (cdr l)))))\n"
            "(define and (b c) (if b  c  b))\n"
            "(define or  (b c) (if b  b  c))\n"
            "(define not (b)   (if b #f #t))\n"

"(define atom? (x) (or (number? x) (or (symbol? x) (or (boolean? x) (null? x)))))\n"
            "(define equal? (s1 s2)\n"
            "  (if (or (atom? s1) (atom? s2))\n"
            "    (= s1 s2)\n"

           "    (and (equal? (car s1) (car s2)) (equal? (cdr s1) (cdr s2)))))\n"
            "(define append (l1 l2)\n"
            "  (if (null? l1)\n"
            "     l2\n"
            "     (cons (car l1) (append (cdr l1) l2))))\n"
            "(define revapp (l1 l2)\n"
            "  (if (null? l1)\n"
            "     l2\n"
            "     (revapp (cdr l1) (cons (car l1) l2))))\n"
            "(define bind (x y alist)\n"
            "  (if (null? alist)\n"
            "    (list1 (list2 x y))\n"
            "    (if (equal? x (caar alist))\n"
            "      (cons (list2 x y) (cdr alist))\n"
            "      (cons (car alist) (bind x y (cdr alist))))))\n"
            "(define find (x alist)\n"
            "  (if (null? alist) '()\n"
            "    (if (equal? x (caar alist))\n"
            "      (car (cdar alist))\n"
            "      (find x (cdr alist)))))\n"
            "(define o (f g) (lambda (x) (f (g x))))\n"
            "(define curry   (f) (lambda (x) (lambda (y) (f x y))))\n"
            "(define uncurry (f) (lambda (x y) ((f x) y)))\n"
            "(define filter (p? l)\n"
            "  (if (null? l)\n"
            "    '()\n"
            "    (if (p? (car l))\n"
            "      (cons (car l) (filter p? (cdr l)))\n"
            "      (filter p? (cdr l)))))\n"
            "(define map (f l)\n"
            "  (if (null? l)\n"
            "    '()\n"
            "    (cons (f (car l)) (map f (cdr l)))))\n"
            "(define exists? (p? l)\n"
            "  (if (null? l)\n"
            "    #f\n"
            "    (if (p? (car l)) \n"
            "      #t\n"
            "      (exists? p? (cdr l)))))\n"
            "(define all? (p? l)\n"
            "  (if (null? l)\n"
            "    #t\n"
            "    (if (p? (car l))\n"
            "      (all? p? (cdr l))\n"
            "      #f)))\n"
            "(define foldr (op zero l)\n"
            "  (if (null? l)\n"
            "    zero\n"
            "    (op (car l) (foldr op zero (cdr l)))))\n"
            "(define foldl (op zero l)\n"
            "  (if (null? l)\n"
            "    zero\n"
            "    (foldl op (op (car l) zero) (cdr l))))\n"
            "(define <= (x y) (not (> x y)))\n"
            "(define >= (x y) (not (< x y)))\n"
            "(define != (x y) (not (= x y)))\n"
            "(define max (x y) (if (> x y) x y))\n"
            "(define min (x y) (if (< x y) x y))\n"
            "(define mod (m n) (- m (* n (/ m n))))\n"
            "(define gcd (m n) (if (= n 0) m (gcd n (mod m n))))\n"
            "(define lcm (m n) (* m (/ n (gcd m n))))\n"
            "(define caar (x) (car (car x)))\n"
            "(define cdar (x) (cdr (car x)))\n"
            "(define cadr (x) (car (cdr x)))\n"
            "(define cddr (x) (cdr (cdr x)))\n"
            "(define caaar (x) (car (caar x)))\n"
            "(define cdaar (x) (cdr (caar x)))\n"
            "(define caadr (x) (car (cadr x)))\n"
            "(define cdadr (x) (cdr (cadr x)))\n"
            "(define cadar (x) (car (cdar x)))\n"
            "(define cddar (x) (cdr (cdar x)))\n"
            "(define caddr (x) (car (cddr x)))\n"
            "(define cdddr (x) (cdr (cddr x)))\n"
            "(define list1 (x)               (cons x '()))\n"
            "(define list2 (x y)             (cons x (list1 y)))\n"
            "(define list3 (x y z)           (cons x (list2 y z)))\n"
            "(define list4 (x y z a)         (cons x (list3 y z a)))\n"
            "(define list5 (x y z a b)       (cons x (list4 y z a b)))\n"
            "(define list6 (x y z a b c)     (cons x (list5 y z a b c)))\n"
            "(define list7 (x y z a b c d)   (cons x (list6 y z a b c d)))\n"
            "(define list8 (x y z a b c d e) (cons x (list7 y z a b c d e)))\n";

        if (setjmp(errorjmp))
            assert(0);  /* fail if error occurs in basis */
        readevalprint(topreader(stringreader("initial basis", basis), 0), &env,
                                                                             0);
    }

    input = topreader(filereader("standard input", stdin), doprompt);

    while (setjmp(errorjmp))
        resetallocate(&env);
    readevalprint(input, &env, 1);
    return 0;
}
