;;;usm.nw:3187
(class Collection Object
  () ; abstract
  (method do:     (aBlock)    (subclassResponsibility self))
  (method add:    (newObject) (subclassResponsibility self))
  (method remove:ifAbsent: (oldObject exnBlock)
                              (subclassResponsibility self))
  (method species ()          (subclassResponsibility self))
  
;;;usm.nw:3199
(classMethod with: (anObject) (locals temp)
    (set temp (new self))
    (add: temp anObject)
    temp)
;;;usm.nw:3211
(method remove: (oldObject) 
    (remove:ifAbsent: self oldObject [(error: self #tried-to-remove-absent-object)]))
(method addAll: (aCollection) 
    (do: aCollection (block (x) (add: self x)))
    aCollection)
(method removeAll: (aCollection) 
    (do: aCollection (block (x) (remove: self x)))
    aCollection)
;;;usm.nw:3231
(method isEmpty () (= (size self) 0))
(method size () (locals temp)
    (set temp 0)
    (do: self (block (_) (set temp (+ temp 1))))
    temp)
(method occurrencesOf: (anObject) (locals temp)
    (set temp 0)
    (do: self (block (x)
       (ifTrue: (= x anObject) [(set temp (+ temp 1))])))
    temp)
(method includes: (anObject) (< 0 (occurrencesOf: self anObject)))
(method detect: (aBlock) 
    (detect:ifNone: self aBlock [(error: self #no-object-detected)]))
(method detect:ifNone: (aBlock exnBlock) (locals answer searching)
    (set searching true)
    (do: self (block (x)
        (ifTrue: (and: searching [(value aBlock x)])
             [(set searching false)
              (set answer x)])))
    (if searching exnBlock [answer]))
;;;usm.nw:3263
(method inject:into: (thisValue binaryBlock)
   (do: self (block (x) (set thisValue (value binaryBlock x thisValue))))
   thisValue)
;;;usm.nw:3271
(method select: (aBlock) (locals temp)
   (set temp (new (species self)))
   (do: self (block (x) (ifTrue: (value aBlock x) [(add: temp x)])))
   temp)
(method reject: (aBlock)
   (select: self (block (x) (not (value aBlock x)))))
(method collect: (aBlock) (locals temp)
   (set temp (new (species self)))
   (do: self (block (x) (add: temp (value aBlock x))))
   temp)
;;;usm.nw:3284
(method asSet () (locals temp)
     (set temp (new Set))
     (do: self (block (x) (add: temp x)))
     temp)
;;;usm.nw:3293
(method print ()
    (printName self)
    (print lparen)
    (do: self (block (x) (print space) (print x)))
    (print space)
    (print rparen)
    self)
(method printName () (print #Collection))
;;;usm.nw:3195
)
;;;usm.nw:3315
(class Set Collection
    (members)  ; list of elements
    (classMethod new () (initSet (new super)))
    (method initSet  () (set members (new List)) self) ; private
    (method do: (aBlock) (do: members aBlock))
    (method add: (item)
        (ifFalse: (includes: members item) [(add: members item)])
        item)
    (method remove:ifAbsent: (item exnBlock) 
        (remove:ifAbsent: members item exnBlock))
    (method asSet     () self)
    (method printName () (print #Set))
    (method species   () Set)
)
;;;usm.nw:3359
(class KeyedCollection Collection
    ()  ; abstract class
    (method at:put: (key value)       (subclassResponsibility self))
    (method associationsDo: (aBlock)  (subclassResponsibility self))
    
;;;usm.nw:3370
(method do: (aBlock) 
    (associationsDo: self (block (x) (value aBlock (value x)))))
(method at: (key)    
    (at:ifAbsent: self key [(error: self #key-not-found])))
(method at:ifAbsent: (key exnBlock) 
    (value (associationAt:ifAbsent: self key 
               [(key:value: Association nil (value exnBlock))])))
(method includesKey: (key) 
    (isKindOf: (associationAt:ifAbsent: self key []) Association))
(method associationAt: (key) 
    (associationAt:ifAbsent: self key [(error: self #key-not-found])))
(method associationAt:ifAbsent: (key exnBlock) (locals finishBlock)
    (set finishBlock exnBlock)
    (associationsDo: self (block (x) 
        (ifTrue: (= (key x) key) [(set finishBlock [x])])))
    (value finishBlock))
;;;usm.nw:3392
(method keyAtValue: (value) 
    (keyAtValue:ifAbsent: self value [(error: self #value-not-found)]))
(method keyAtValue:ifAbsent: (value aBlock) (locals finishBlock)
    (set finishBlock aBlock)
    (associationsDo: self (block (x) 
        (ifTrue: (= (value x) value) [(set finishBlock [(key x)])])))
    (value finishBlock))
;;;usm.nw:3364
)
;;;usm.nw:3433
(class Association Object 
   (key value)
   (classMethod key:value: (a b) (setKey:value: (new self) a b))
   (method setKey:value: (x y) (set key x) (set value y) self) ; private
   (method key    ()  key)
   (method value  ()  value)
   (method key:   (x) (set key   x))
   (method value: (y) (set value y))
)
;;;usm.nw:3446
(class Dictionary KeyedCollection
    (table) ; list of pairs
    (classMethod new ()       (initDictionary (new super)))
    (method initDictionary () (set table (new List)) self) ; private
    (method printName ()      (print #Dictionary))
    (method species ()        Dictionary)
    
;;;usm.nw:3460
(method associationsDo: (aBlock) (do: table aBlock))
(method at:put: (key value) (locals tempassoc)
    (set tempassoc (associationAt:ifAbsent: self key []))
    (if (isNil tempassoc)
         [(add: table (key:value: Association key value))]
         [(value: tempassoc value)])
    value)
;;;usm.nw:3471
(method add: (_) (error: self #can't-just-add:-to-a-Dictionary))
;;;usm.nw:3453
)
;;;usm.nw:3484
(class SequenceableCollection KeyedCollection
    () ; abstract class
    (method firstKey () (subclassResponsibility self))
    (method lastKey  () (subclassResponsibility self))
    (method last     () (at: self (lastKey  self)))
    (method first    () (at: self (firstKey self)))
    (method at:ifAbsent: (index exnBlock) (locals current resultBlock)
        (set resultBlock exnBlock)
        (set current (firstKey self))
        (do: self (block (v)
            (ifTrue: (= current index) [(set resultBlock [v])])
            (set current (+ current 1))))
        (value resultBlock))
)
;;;usm.nw:3906
(class Cons Object
    (car cdr)
    (method car ()           car)
    (method cdr ()           cdr)
    (method car: (anObject)  (set car anObject) self)
    (method cdr: (anObject)  (set cdr anObject) self)
    
;;;usm.nw:3921
(method deleteAfter () (locals answer)
    (set answer (car cdr))
    (set cdr    (cdr cdr))
    (pred: cdr self)
    answer)
(method insertAfter: (anObject)
    (set cdr (car: (cdr: (new Cons) cdr) anObject))
    (pred: (cdr cdr) cdr)
    anObject)
;;;usm.nw:3942
(method do: (aBlock)
    (value aBlock car)
    (do: cdr aBlock))
;;;usm.nw:3952
(method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)
    (if (value aBlock self)
        [(deleteAfter pred)]
        [(rejectOne:ifAbsent:withPred: cdr aBlock exnBlock self)]))
;;;usm.nw:3913
    (method pred: (aCons)    nil)
)
;;;usm.nw:3963
(class ListSentinel Cons
    (pred)
    (method do:   (aBlock)  nil)
    (method rejectOne:ifAbsent:withPred: (aBlock exnBlock pred)
                            (value exnBlock))
    (method pred: (aCons)   (set pred aCons))
    (method pred  ()        pred)
    (classMethod new ()     (locals tmp)
        (set tmp (new super))
        (pred: tmp tmp)
        (cdr:  tmp tmp)
        tmp))
;;;usm.nw:3833
(class List SequenceableCollection
    (sentinel)
    (classMethod new  ()        (sentinel: (new super) (new ListSentinel)))
    (method sentinel: (s)       (set sentinel s) self) ; private
    (method isEmpty   ()        (= sentinel (cdr sentinel)))
    
;;;usm.nw:3852
(method addLast:  (item)   (insertAfter: (pred sentinel) item))
(method addFirst: (item)   (insertAfter: sentinel item))
(method add: (item)        (addLast: self item))
;;;usm.nw:3859
(method removeFirst ()     (deleteAfter sentinel))
;;;usm.nw:3873
(method remove:ifAbsent: (oldObject exnBlock)
    (rejectOne:ifAbsent:withPred:
        (cdr sentinel)
        (block (x) (= oldObject (car x)))
        exnBlock
        sentinel))
;;;usm.nw:3886
(method firstKey () 1)
(method lastKey  () (size self))
(method at:put: (n value) (locals tmp)
    (set tmp (cdr sentinel))
    (whileFalse: [(= n 1)]
       [(set n (- n 1))
        (set tmp (cdr tmp))])
    (car: tmp value))
;;;usm.nw:3839
    (method last      ()        (car (pred sentinel)))
    (method do:       (aBlock)  (do: (cdr sentinel) aBlock))
    (method species   ()        List)
    (method printName ()        (print #List))
)
;;;usm.nw:4040
(class Array SequenceableCollection
    () ; representation is primitive
    (classMethod new: primitive arrayNew:)
    (classMethod new () (error: self #size-of-Array-must-be-specified))
    (method size     primitive arraySize)
    (method at:      primitive arrayAt:)
    (method at:put:  primitive arrayAt:Put:)
    
;;;usm.nw:4057
(method add:             (x)   (fixedSizeError self))
(method remove:ifAbsent: (x b) (fixedSizeError self))
(method fixedSizeError   ()    (error: self #arrays-have-fixed-size))
;;;usm.nw:4064
(method select:  (_) (error: self #select-on-arrays-not-implemented))
(method collect: (_) (error: self #collect-on-arrays-not-implemented))
;;;usm.nw:4070
(method firstKey () 1)
(method lastKey  () (size self))
(method do: (aBlock) (locals index)
    (set index (firstKey self))
    (timesRepeat: (size self)
       [(value aBlock (at: self index))
        (set index (+ index 1))]))
;;;usm.nw:4048
    (method species   () Array)
    (method printName () nil) ; names of arrays aren't printed
)
