;;;usm.nw:4106
(class Magnitude Object 
    () ; abstract class
    (method =  (x) (subclassResponsibility self)) ; may not inherit from Object
    (method <  (x) (subclassResponsibility self))
    (method >  (y) (< y self))
    (method <= (x) (not (> self x)))
    (method >= (x) (not (< self x)))
    (method min: (aMagnitude) (if (< self aMagnitude) [self] [aMagnitude]))
    (method max: (aMagnitude) (if (> self aMagnitude) [self] [aMagnitude]))
)
;;;usm.nw:4120
(class Number Magnitude
    ()  ; abstract class
    ;;;;;;; basic Number protocol
    (method +   (aNumber)     (subclassResponsibility self))
    (method *   (aNumber)     (subclassResponsibility self))
    (method negated    ()     (subclassResponsibility self))
    (method reciprocal ()     (subclassResponsibility self))
    
    (method asInteger  ()     (subclassResponsibility self))
    (method asFraction ()     (subclassResponsibility self))
    (method asFloat    ()     (subclassResponsibility self))
    (method coerce: (aNumber) (subclassResponsibility self))
    
;;;usm.nw:4138
(method -   (y) (+ self (negated  y)))
(method abs ()  (if (negative self) [(negated  self)] [self]))
(method /   (y) (* self (reciprocal y)))
;;;usm.nw:4146
(method negative         () (<  self (coerce: self 0)))
(method positive         () (>= self (coerce: self 0)))
(method strictlyPositive () (>  self (coerce: self 0)))
;;;usm.nw:4154
(method squared () (* self self))
(method raisedToInteger: (anInteger)
    (if (= anInteger 0)
        [(coerce: self 1)]
        [(if (= anInteger 1) [self]
            [(* (squared (raisedToInteger: self (div: anInteger 2)))
                (raisedToInteger: self (mod: anInteger 2)))])]))
;;;usm.nw:4181
(method sqrt () (sqrtWithin: self (coerce: self (/ 1 10))))
(method sqrtWithin: (epsilon) (locals two x_{i-1} x_{i})
    ; find square root of receiver within epsilon
    (set two     (coerce: self 2))
    (set x_{i-1} (coerce: self 1))
    (set x_{i}   (/ (+ x_{i-1} (/ self x_{i-1})) two))
    (while [(> (abs (- x_{i-1} x_{i})) epsilon)]
           [(set x_{i-1} x_{i})
            (set x_{i} (/ (+ x_{i-1} (/ self x_{i-1})) two))])
    x_{i})
;;;usm.nw:4133
)
;;;usm.nw:4513
(class Integer Number
    () ; abstract class
    (method div: (n) (subclassResponsibility self))
    (method mod: (n) (- self (* n (div: self n))))
    (method gcd: (n) (if (= n (coerce: self 0)) [self] [(gcd: n (mod: self n))]))
    (method lcm: (n) (* self (div: n (gcd: self n))))
    
;;;usm.nw:4529
(method asFraction () (num:den:  Fraction self 1))
(method asFloat    () (mant:exp: Float    self 0))
(method asInteger  () self)
;;;usm.nw:4538
(method coerce: (aNumber) (asInteger aNumber))
;;;usm.nw:4543
(method reciprocal () (num:den: Fraction 1 self)) 
(method / (aNumber) (/ (asFraction self) aNumber))
;;;usm.nw:4549
(method timesRepeat: (aBlock) (locals count)
    (ifTrue: (negative self) [(error: self #negative-repeat-count)])
    (set count self)
    (while [(!= count 0)]
         [(value aBlock)
          (set count (- count 1))]))
;;;usm.nw:4520
)
;;;usm.nw:4576
(class SmallInteger Integer
    () ; primitive representation
    (classMethod new: primitive newSmallInteger:)
    (classMethod new   () (new: self 0))
    (method negated    () (- 0 self))
    (method print primitive printSmallInteger)
    (method +     primitive +)
    (method -     primitive -)
    (method *     primitive *)
    (method div:  primitive div)
    (method =     primitive eqObject)
    (method <     primitive <)
    (method >     primitive >)
)
;;;usm.nw:4608
(class Fraction Number
    (num den)
    (classMethod num:den: (a b) (initNum:den: (new self) a b))
    (method initNum:den: (a b) ; private
        (setNum:den: self a b)
        (signReduce self)
        (divReduce self))
    (method setNum:den: (a b) (set num a) (set den b) self) ; private
    
;;;usm.nw:4641
(method signReduce () ; private
    (ifTrue: (negative den)
        [(set num (negated num)) (set den (negated den))])
    self)

(method divReduce () (locals temp) ; private
    (if (= num 0)
        [(set den 1)]
        [(set temp (gcd: (abs num) den))
         (set num  (div: num temp))
         (set den  (div: den temp))])
    self)
;;;usm.nw:4660
(method num () num)
(method den () den)
;;;usm.nw:4669
(method = (f)
    (and: (= num (num f)) [(= den (den f))]))
(method < (f)
    (< (* num (den f)) (* (num f) den)))
;;;usm.nw:4678
(method negated () (setNum:den: (new Fraction) (negated num) den))
;;;usm.nw:4692
(method * (f)
    (divReduce
        (setNum:den: (new Fraction)
                        (* num (num f))
                        (* den (den f)))))
;;;usm.nw:4711
(method + (f) (locals temp)
    (set temp (lcm: den (den f)))
    (divReduce
        (setNum:den: (new Fraction)
                     (+ (* num (div: temp den)) (* (num f) (div: temp (den f))))
                     temp)))
;;;usm.nw:4725
(method reciprocal ()
   (signReduce (setNum:den: (new Fraction) den num)))
;;;usm.nw:4730
(method print () (print num) (print #/) (print den) self)
;;;usm.nw:4740
(method asInteger  () (div: num den))
(method asFloat    () (/ (asFloat num) (asFloat den)))
(method asFraction () self)
(method coerce: (aNumber) (asFraction aNumber))
;;;usm.nw:4749
(method negative         () (negative num))
(method positive         () (positive num))
(method strictlyPositive () (strictlyPositive num))
;;;usm.nw:4617
)
;;;usm.nw:4782
(class Float Number
    (mant exp)
    (classMethod mant:exp: (m e) (initMant:exp: (new self) m e))
    (method initMant:exp: (m e) ; private
        (set mant m) (set exp e) (normalize self))
    (method normalize ()    ; private
        (while [(> (abs mant) 32767)]
               [(set mant (div: mant 10))
                (set exp (+ exp 1))])
        self)
    
;;;usm.nw:4798
(method mant () mant)  ; private
(method exp  () exp)   ; private
;;;usm.nw:4806
(method < (x) (negative (- self x)))
(method = (x) (isZero   (- self x)))
(method isZero () (= mant 0))
;;;usm.nw:4813
(method negated () (mant:exp: Float (negated mant) exp))
;;;usm.nw:4834
(method + (prime) 
    (if (>= exp (exp prime))
        [(mant:exp: Float (+ (* mant (raisedToInteger: 10 (- exp (exp prime))))
                             (mant prime))
                          (exp prime))]
        [(+ prime self)]))
;;;usm.nw:4846
(method * (prime) 
    (mant:exp: Float (* mant (mant prime)) (+ exp (exp prime))))
;;;usm.nw:4855
(method reciprocal ()
    (mant:exp: Float (div: 1000000000 mant) (- -9 exp)))
;;;usm.nw:4861
(method coerce: (aNumber) (asFloat aNumber))
(method asFloat () self)
;;;usm.nw:4867
(method asInteger ()
    (if (< exp 0)
        [(div: mant (raisedToInteger: 10 (negated exp)))]
        [(*    mant (raisedToInteger: 10 exp))]))
;;;usm.nw:4875
(method asFraction ()
    (if (< exp 0)
        [(num:den: Fraction mant (raisedToInteger: 10 (negated exp)))]
        [(num:den: Fraction (* mant (raisedToInteger: 10 exp)) 1)]))
;;;usm.nw:4896
(method negative         () (negative mant))
(method positive         () (positive mant))
(method strictlyPositive () (strictlyPositive mant))
;;;usm.nw:4904
(method print () 
    (print-normalize self) 
    (print mant) (print #x10^) (print exp)
    self)

(method print-normalize ()
    (while [(and: (< exp 0) [(= (mod: mant 10) 0)])]
        [(set exp (+ exp 1)) (set mant (div: mant 10))]))
;;;usm.nw:4793
)
