;;;usm.nw:2207
(class Simulation Object
     () ; abstract class
     (method initialize () (subclassResponsibility self))
     (method report ()     (subclassResponsibility self))
     (method runUntil: (timelimit)
         (set CLOCK 0)
         (set EVQUEUE (new EventQueue))
         (initialize self)
         (while [(<= CLOCK timelimit)] [(doNextEvent EVQUEUE)])
         (report self))
)
(val CLOCK   nil)
(val EVQUEUE nil)
;;;usm.nw:2237
(class LabSimulation Simulation
    (termlimit)
    (classMethod withLimit: (t) (termlimit: (new self) t))
    (method termlimit: (t)
        (set termlimit t) self)
    (method initialize ()
        (set TERMINALLIMIT     termlimit)
        (set THELAB            (new Lab))
        (set STUDENTNO         0)
        (set STUDENTSFINISHING 0)
        (set TOTALTIMEWASTED   0)
        (set TERMINALQUEUE     (new Queue))
        (set WAITTIMES         (new WaitTimeList))
        (set SERVICETIMES      (new ServiceTimeList))
        (scheduleNewArrival (new Student)))
    (method report ()
        (println #simulation-done)
        (print   #students-finishing=)
        (println STUDENTSFINISHING)
        (print   #left-on-queue=)
        (println (size TERMINALQUEUE))
        (print   #Total-time-wasted=)
        (println TOTALTIMEWASTED)
        (print   #Average-time-wasted=)
        (println (div: TOTALTIMEWASTED STUDENTSFINISHING))
        self)
)

(val TERMINALLIMIT     nil)
(val THELAB            nil)
(val STUDENTNO         nil)
(val STUDENTSFINISHING nil)
(val TOTALTIMEWASTED   nil)
(val TERMINALQUEUE     nil)
(val WAITTIMES         nil)
(val SERVICETIMES      nil)
;;;usm.nw:2330
(class Lab Object
    (term1free term2free)
    (classMethod new () (initLab (new super)))
    (method initLab ()
        (set term1free true)
        (set term2free true)
        self)
    (method terminals-free? () (| term1free term2free))
    (method pick-terminal ()
         (if term1free
              [(set term1free false) 1]
              [(set term2free false) 2]))
    (method release-terminal: (t)
         (if (= t 1) [(set term1free true)] [(set term2free true)]))
)
;;;usm.nw:2357
(class Queue List
    ()
    (method species     () Queue)
    (method printName   () (print #Queue))
)
;;;usm.nw:2385
(class EventQueue Object
    (pqueue) ; a PriorityQueue containing event objects
    (classMethod new () (initEventQueue (new super)))
    (method initEventQueue () ; private
         (set pqueue (new PriorityQueue)) self)
    (method scheduleEvent:at: (event time)
         (addAssociation: pqueue (key:value: Association time event)))
    (method doNextEvent () (locals pair)
           (set pair  (deleteMin pqueue))
           (set CLOCK (key pair))
           (takeAction (value pair)))
)
;;;usm.nw:2419
(class PriorityQueue List
    () ; sorted list of associations
    (method deleteMin () (removeFirst self))
    (method bubble () ; bubble first element to its proper place (private)
        (locals tmp p) 
        (car: sentinel (car (cdr sentinel)))
        (set p (cdr sentinel))
        (whileTrue: [(> (key (car p)) (key (car (cdr p))))]
            [(set tmp (car p))
             (car: p (car (cdr p)))
             (car: (cdr p) tmp)
             (set p (cdr p))]))
    (method addAssociation: (anAssociation)
        (addFirst: self anAssociation)
        (bubble self))
    (method species   () PriorityQueue)
    (method printName () (print #PriorityQueue))
)
;;;usm.nw:2466
(class Student Object
    (status number timeNeeded timeStillNeeded arrivalTime)
    (method print () (print #<Student) (print space) (print number) (print #>))
    (method setNumber:time:arrival: (n t a)
         (set status          #scheduled)
         (set number          n)
         (set timeNeeded      t)
         (set timeStillNeeded t)
         (set arrivalTime     a)
         self)
    
;;;usm.nw:2519
(method takeAction ()
     (if (= status #scheduled) [(arrive self)] [(leaveTerminal self)]))
;;;usm.nw:2528
(method arrive ()
     (if (terminals-free? THELAB)
          [(grabTerminal self)]
          [(set status #on-queue)
           (addLast: TERMINALQUEUE self)])
     (scheduleNewArrival (new Student)))
;;;usm.nw:2547
(method leaveTerminal () (locals wasted)
     (release-terminal: THELAB status)
     (if (> timeStillNeeded 0)
          [(set status #on-queue)
           (addLast: TERMINALQUEUE self)
           (grabTerminal (removeFirst TERMINALQUEUE))]

          [(set status            #finished)
           (set wasted            (- (- CLOCK arrivalTime) timeNeeded))
           (set STUDENTSFINISHING (+ 1 STUDENTSFINISHING))
           (set TOTALTIMEWASTED   (+ TOTALTIMEWASTED wasted))
           (ifFalse: (isEmpty TERMINALQUEUE)  ; terminal to next student, if any
               [(grabTerminal (removeFirst TERMINALQUEUE))])]))
;;;usm.nw:2564
(method grabTerminal ()
     (set status (pick-terminal THELAB)) ; using a terminal
     (scheduleLeaveTerminal self))
;;;usm.nw:2592
(method scheduleLeaveTerminal ()
     (if (<= timeStillNeeded TERMINALLIMIT)
          [(scheduleEvent:at: EVQUEUE self (+ CLOCK timeStillNeeded))
           (set timeStillNeeded 0)]

          [(scheduleEvent:at: EVQUEUE self (+ CLOCK TERMINALLIMIT))
           (set timeStillNeeded (- timeStillNeeded TERMINALLIMIT))]))
;;;usm.nw:2614
(method scheduleNewArrival () (locals wait service arrival)
     (set wait      (next WAITTIMES))
     (set service   (next SERVICETIMES))
     (set STUDENTNO (+ 1 STUDENTNO))
     (set arrival   (+ CLOCK wait))
     (setNumber:time:arrival: self STUDENTNO service arrival)
     (scheduleEvent:at: EVQUEUE self arrival))
;;;usm.nw:2477
)
;;;usm.nw:2680
; Twenty arrivals at time zero
(class WaitTimeList Object
    (which)
    (classMethod new () (initWaitTimeList (new super)))
    (method initWaitTimeList () (set which 0) self)
    (method next ()
         (if (= which 20) [2000] [(set which (+ 1 which)) 0]))
)
; All students need 120 minutes of terminal time
(class ServiceTimeList Object
    ()
    (classMethod new () (initServiceTimeList (new super)))
    (method initServiceTimeList () self)
    (method next () 120)
)
